/*
 * HAL.c
 *
 * Created: 23.12.2019 14:03:11
 *  Author: GoldenAndy
 */ 

#include "HAL.h"
#include <util/delay.h>

//////////////////////////////////////////////////////////////////////////
// I2C

	#define USED_PORTC_IIC		( _BV(PORTC4) | _BV(PORTC5) )

	static void i2cInit(void){
		DDRC &= ~( _BV(PORTC4) | _BV(PORTC5) ); // 
		PORTC &= ~( _BV(PORTC4) | _BV(PORTC5) ); //  .    .
		#if (F_CPU == 16000000UL) 
			iicInit( 0x0C, 0x00 ); // 16 MHz
		# else
			iicInit( 0x02, 0x00 ); // 8 MHz	- 0x02 - 400, 0x0C - 200 kHz
		#endif
	}
	
	//------------------------------------------- 
	static void displaySearchAndInit(void) {
		if (ssd1306init(OLED_SSD1306_ADDR1) ) return;
		if (ssd1306init(OLED_SSD1306_ADDR2) ) return;
		while (1) {} //    . .
	}
	
	//-------------------------------------------    i2c  
	uint8_t i2c_sendSSD1306data(uint8_t i2cAddr8bit, uint8_t prefix, uint8_t* buffer, uint8_t dataCount){
		if ( iicSetStart() == iicError ) {
	stopAndError:		
			iicSetStop();
			return 0;
		}
		if (iicSendAddr(i2cAddr8bit, iicWriteData) != iicOkACK) goto stopAndError;
		if (iicSendData(prefix) != iicOkACK) goto stopAndError;
		uint8_t n = 0;
		while (n < dataCount) {
			if (iicSendData(buffer[n]) != iicOkACK) goto stopAndError;
			n++;
		} // while
		iicSetStop();
		return 1;
	}
	
	//-------------------------------------------    i2c  
	uint8_t i2c_WriteAddrWriteData(uint8_t i2c_7bitAddr, uint8_t Reg, void* Value, uint8_t dataCount){
		//    ,   
		//      0,  1
		return i2c_sendSSD1306data( i2c_7bitAddr * 2, Reg, Value, dataCount ); 
	}
	
	//------------------------------------------- 
	uint8_t i2c_WriteAddrReadData(uint8_t i2c_7bitAddr, uint8_t Reg, void* Value, uint8_t dataCount){
		uint8_t i2c_8bitAddr = i2c_7bitAddr * 2;
		//    ,   
		//      0,  1
		if ( iicSetStart() == iicError ) {
	stopAndError:
			iicSetStop();
			return 0;
		}
		if ( iicSendAddr(i2c_8bitAddr, iicWriteData) != iicOkACK ) goto stopAndError;
		if ( iicSendData(Reg) != iicOkACK )						   goto stopAndError;
		if ( iicSetStart() == iicError )						   goto stopAndError;
		if ( iicSendAddr(i2c_8bitAddr, iicReadData) != iicOkACK )  goto stopAndError;
		uint8_t n = 0;
		uint8_t* buffer = Value;
		while (n < dataCount) {
			if (n<dataCount-1) {
				//    ,   ACK
				if (iicReceiveData(iicACKflag) != iicOkACK ) goto stopAndError;
				buffer[n] = iisGetReceivedData();
			} else {
				//   ,   NACK
				if (iicReceiveData(iicNACKflag) != iicOkNACK ) goto stopAndError;
				buffer[n] = iisGetReceivedData();
			}
			n++;
		} // while
		iicSetStop();
		return 1;
	}

//////////////////////////////////////////////////////////////////////////
// 
	#define KEYS_PORT			D
	#define USED_PORTD_KEYS		KEYS_MASK
	
	//------------------------------------------- init 
	static void keysInit(void){
		__DDRPORT(KEYS_PORT) &= ~KEYS_MASK; // input
		__OUTPORT(KEYS_PORT) |=  KEYS_MASK; // pull_up
	}
	//------------------------------------------- get keys state 
	uint8_t keysGetPress(void) {
		return ( ~__INPORT(KEYS_PORT) ) & KEYS_MASK;
	}
	//------------------------------------------- wake up interrupt
	void keysEnableWakeUp(uint8_t wakeUpMask){
		// Port D -> PCICR(PCIE2), PCIFR(PCIF2), PCMSK2, PCINT2_vect
		PCIFR |= _BV(PCIF2);	// clear flag
		PCICR |= _BV(PCIE2);	// enable interrupt for [16..23] group
		PCMSK2 = KEYS_MASK;		// enable interrupt from our keys
	}
	
	void keysDisableWakeUp(void){
		PCIFR |=  _BV(PCIF2);	// clear flag
	} 

//////////////////////////////////////////////////////////////////////////
// Phones
	#define PHONES_PORT			D
	#define PHONES_BITVAL		_BV(PORTD2)
	#define USED_PORTD_PHONES	PHONES_BITVAL

	static uint8_t phonesCheckPresentMode = 0;

	void phonesEnableOutput(void){
		//  
		phonesCheckPresentMode = 0;
		__OUTPORT(PHONES_PORT) &= ~PHONES_BITVAL;
		__DDRPORT(PHONES_PORT) |=  PHONES_BITVAL;
	}

	void phonesEnableCheckPresent(void){
		//    
		phonesCheckPresentMode = 1;
		__DDRPORT(PHONES_PORT) &= ~PHONES_BITVAL;
		__OUTPORT(PHONES_PORT) |=  PHONES_BITVAL;
	}
					
	void phonesDisableCheckPresent(void){
		//    
		phonesCheckPresentMode = 0;
		__DDRPORT(PHONES_PORT) |= PHONES_BITVAL;
		__OUTPORT(PHONES_PORT) &= ~PHONES_BITVAL;
	}
	
	void phonesSetOut(uint8_t outValue){
		//      
		if (outValue) __OUTPORT(PHONES_PORT) |= PHONES_BITVAL;
		else __OUTPORT(PHONES_PORT) &= ~PHONES_BITVAL;
	}
	
	uint8_t phonesGetPresent(void){
		//     (     )
		if ( !phonesCheckPresentMode ) return 1;
		return (~__INPORT(PHONES_PORT)) & PHONES_BITVAL;
	}

//////////////////////////////////////////////////////////////////////////
// Ext. amp control
	#define EXTAMP_PORT			B
	#define EXTAMP_BITVAL		_BV(PORTB5)
	#define USED_PORTB_EXTAMP	EXTAMP_BITVAL
	
	void extAmpControlSetState(uint8_t state){
		__DDRPORT(EXTAMP_PORT) |= EXTAMP_BITVAL;
		if (state) __OUTPORT(EXTAMP_PORT) &= ~EXTAMP_BITVAL;
		else __OUTPORT(EXTAMP_PORT) |= EXTAMP_BITVAL;
	}

//////////////////////////////////////////////////////////////////////////
//   T2
	static void secondsTimerInit(void){
		//   32768  
		// 
		ASSR = _BV(AS2);
		TCCR2A = 0;
		TCCR2B = ( 1 << CS22 | 0 << CS21 | 1 << CS20 ); // CS2[2:0] = 101 - prescaler = 128
		TCNT2 = 0;
		while (ASSR & ( _BV(TCN2UB) | _BV(OCR2AUB) | _BV(OCR2BUB) | _BV(TCR2AUB) | _BV(TCR2BUB) ) ) {}
		TIFR2 |= _BV(TOV2);
		TIMSK2 = _BV(TOIE2);
	}

//////////////////////////////////////////////////////////////////////////
//  
	//-------------------------------------------  
	#define CHARGING_PORT		B
	#define CHARGING_BITVAL		_BV(PORTB0)
	#define USED_PORTB_CHARGER	CHARGING_BITVAL
	
	static void chargingInit(void){
		__DDRPORT(CHARGING_PORT) &= ~CHARGING_BITVAL;
		__OUTPORT(CHARGING_PORT) |=  CHARGING_BITVAL;
	}
	
	uint8_t batteryGetChargingState(void){
		//   
		return ( ~__INPORT(CHARGING_PORT)) & CHARGING_BITVAL;
	}

	#define MEASURE_CYCLES			16
	//-------------------------------------------  
	uint16_t getAdcValue(uint8_t dummyMeas){
		uint16_t adcRes = MEASURE_CYCLES / 2;
		for (uint8_t i = 0 ; i < MEASURE_CYCLES; i++) {
			ADCSRA |= ( 1 << ADSC );
			while ( ! (ADCSRA & ( 1 << ADIF )) ) {}
			ADCSRA |= ( 1 << ADIF );
			adcRes += ADC;
			if (dummyMeas) return 0;
		}
		return adcRes/MEASURE_CYCLES;
	}

	//-------------------------------------------  
	static void enableAdc(void) {
		//   BANDGAP: MUX[3:0] = 1110,   ,  Reference: REFS[1:0] = 01
		ADMUX = ( 0 << REFS1 ) | ( 1 << REFS0 ) | /*AVCC is VRef*/\
				( 0 << ADLAR ) | \
				( 1 << MUX3 ) | ( 1 << MUX2 ) | ( 1 << MUX1 ) | ( 0 << MUX0 )/*1110 - input = VBandGap*/;
	#	if (F_CPU == 16000000UL)
		ADCSRA = ( 1 << ADEN ) | ( 0 << ADSC ) | ( 1 << ADIF ) | \
				 ( 1 << ADPS2 ) | ( 1 << ADPS1 ) | ( 0 << ADPS0 ); /* 110 - divider = 64 */
	#	else
		ADCSRA = ( 1 << ADEN ) | ( 0 << ADSC ) | ( 1 << ADIF ) | \
				 ( 1 << ADPS2 ) | ( 0 << ADPS1 ) | ( 1 << ADPS0 ); /* 101 - divider = 32 */
	#	endif
		_delay_us(60);//    
		//  
		getAdcValue(1);
	}
	
	//-------------------------------------------   
	static void disableAdc(void){
		ADMUX = ( 0 << REFS1 ) | ( 1 << REFS0 ); //   Vref  Vcc
		ADCSRA = 0;
	}
	
	//-------------------------------------------   
	
	uint16_t batteryGetVoltage(uint8_t vBandgap){
		/************************************
		 *	   ADCres = Vin * 1024 / Vref
		 *	  Vin = Vbandgap (~1.10 ),    Vref
		 *	Vref = Vbandgap * 1024 / ADCres
		 *	 Vbandgap   -     
		 *	 .   Vbandgap  
		 *	,       
		 *	       
		 *	 150..200   550  .
		 ************************************/
		enableAdc();
		uint16_t adcRes = getAdcValue(0);
		disableAdc();
		return ( (uint32_t)vBandgap * 1024 ) / adcRes;
	}

	//-------------------------------------------  
	uint8_t batteryGetReferenceVoltage(uint16_t vBat) {
		/************************************
		 *	  .   ADCres = Vin * 1024 / Vref
		 *	  Vref -  vBat -    
		 *	Vin -   bandgap.
		 *	, bandgap = vBat * ADCres / 1024
		 *         vBat
		 *  vBat           
		 ************************************/
		enableAdc();
		uint16_t adcRes = getAdcValue(0);
		disableAdc();
		uint16_t result = ( (uint32_t)adcRes * vBat) / 1024;
		if (result>255) result = 255;
		return result;
	}

//////////////////////////////////////////////////////////////////////////
// System clock
	static void initSysClock(void){
		TCCR0A = ( 1 << WGM01 ) | ( 0 << WGM00 ); // WGM0[2:0] = 010 - CTC mode, top in OCR0A
		TCCR0B = ( 0 << WGM02 ) | ( 0 << CS02 ) | ( 1 << CS01 ) | ( 0 << CS00 );// CS0[2:0] = 010 - divider 8
		OCR0A = 250-1; // 8 MHz / (8*250) = 4kHz
		TIMSK0 |= 1 << OCIE0A;
	}
	
//////////////////////////////////////////////////////////////////////////
// Deep sleep
	void goSleep(void){
		disableAdc();
		SMCR &= ~( (1<<SM2) | (1<<SM1) | (1<<SM0) );// Clear SM[2:0] bits
		SMCR |= (1<<SE) | (0<<SM2) | (1<<SM1) | (1<<SM0);// SM[2:0]=011 - PowerSave
		PRR |= _BV(PRTWI) | _BV(PRTIM1) | _BV(PRSPI) | _BV(PRUSART0) | _BV(PRADC);
		asm volatile ("sleep" : : ); 
		SMCR &= ~(1<<SE);
		PRR &= ~(_BV(PRTWI) | _BV(PRTIM1) | _BV(PRSPI) | _BV(PRUSART0) | _BV(PRADC));
		i2cInit();
		
		
	}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Hardware init
	void hardwareInit(void){
		i2cInit();
		displaySearchAndInit();
		keysInit();
		keysDisableWakeUp();
		phonesDisableCheckPresent();
		chargingInit();
		extAmpControlSetState(0);
		secondsTimerInit();
		disableAdc();
		initSysClock();
		//
		// comparator off
		ACSR |= _BV(ACD);
		// Pull up unused lines of Port B 
		DDRB &=   ( USED_PORTB_CHARGER | USED_PORTB_EXTAMP );
		PORTB |= ~( USED_PORTB_CHARGER | USED_PORTB_EXTAMP );
		// Pull up unused lines of Port C
		DDRC &=   ( USED_PORTC_IIC );
		PORTC |= ~( USED_PORTC_IIC );
		// Pull up unused lines of Port D
		DDRD &=   ( USED_PORTD_KEYS | USED_PORTD_PHONES );
		PORTD |= ~( USED_PORTD_KEYS | USED_PORTD_PHONES );
	}

#undef KEYS_PORT
#undef KEYS_MASK
#undef PHONES_PORT
#undef PHONES_BITVAL
#undef CHARGING_PORT
#undef CHARGING_BITVAL
#undef EXTAMP_PORT
#undef EXTAMP_BITVAL
#undef USED_PORTB_EXTAMP
#undef USED_PORTB_CHARGER
#undef USED_PORTC_IIC
#undef USED_PORTD_KEYS
#undef USED_PORTD_PHONES
#undef USED_PORTB_CHARGER
#undef USED_PORTB_EXTAMP
#undef USED_PORTC_IIC
#undef USED_PORTD_KEYS
#undef USED_PORTD_PHONES
#undef MEASURE_CYCLES
