/*
 * graphics.c
 *
 * Created: 29.12.2019 18:26:37
 *  Author: GoldenAndy
 */ 

#include "graphics.h"
#include "graphicsRes.h"
#include "menu.h"
#include "main.h"
#include "../_common/RDA5807M/RDA5807M.h"
#include "../_common/random/random.h"


//  
static uint8_t longProcessEnd = 0;
// 
static modeType oldMode = 0xFF;
static uint8_t displayPower = 1;
#ifdef USE_CONTRAST_SHIFT
static uint16_t contrastCounter = 0;
#endif
static uint8_t sleepTimeVal = 0;
// 
static uint8_t volumeVal = 0x00;
static uint16_t freqVal = 0x00;
static uint8_t levelVal = 0x00;
static uint8_t stereoVal = 0xFF;
static uint8_t channelVal = 0xFF;
static uint8_t scanChVal = 0xFF;
static char rdsStation[RDS_NAME_LENGTH+1] = {0};
// 
static uint8_t hourVal = 0x00;
static uint8_t minVal = 0x00;
// 
static uint8_t chargeVal = 0x00;
static uint16_t batteryVal = 0x00;
// 
static uint16_t screenEditVal = 0;


//  
typedef enum {
	ufVolume		= 0x01,
	ufFrequency		= 0x02,
	ufTime			= 0x04,
	ufReceiverFlags	= 0x08,
	ufBattery		= 0x10,
	ufChannel		= 0x20,
	ufMenuChange	= 0x40,
	ufRdsStation	= 0x80
} updateFlagsType;

static updateFlagsType updateFlags = 0;

//------------------------------------------- 

#ifdef USE_CONTRAST_SHIFT
#define  MAX_CONTRAST_TIME	4500
void displaySetMaxContrast(void){
	//contrastCounter = MAX_CONTRAST_TIME + 1;
}
#endif

//-------------------------------------------  
static uint8_t int2charArr(const uint16_t value, uint8_t minSymbols, uint8_t shift, char* arr){
	uint8_t n = 1;
	if (value > 9) n = 2;
	if (value > 99) n = 3;
	if (value > 999) n = 4;
	if (value > 9999) n = 5;
	if (n < minSymbols) n = minSymbols;
	uint16_t val = value;
	uint8_t count = n;
	while (n) {
		n--;
		uint16_t x10 = val/10;
		arr[n] =  val - 10 * x10 + shift;
		val = x10;
	}
	return count;
}

static uint8_t int2Frequency(const uint16_t freqVal, char* arr, uint8_t addMHz){
	int2charArr(freqVal, 5, '0', arr);
	arr[5] = arr[4];
	arr[4] = arr[3];
	arr[3] = '.';
	if (arr[0] == '0') arr[0]=' ';
	if (!addMHz) return 6;
	arr[6]=' ';
	arr[7]='M';
	arr[8]='H';
	arr[9]='z';
	return 10;
}

static uint8_t makeTime(char* arr){
	if (hourVal < 0x10 ) arr[0] = ' ';
	else arr[0] = '0' + (hourVal >> 4);
	arr[1] = '0' + (hourVal & 0x0F);
	arr[2] = ':';
	arr[3] = '0' + (minVal >> 4);
	arr[4] = '0' + (minVal & 0x0F);
	return 5;
}

static uint8_t makeBattery(uint16_t showBatVal, char* arr){
	int2charArr(showBatVal, 3, '0', arr);
	arr[3] = arr[2];
	arr[2] = arr[1];
	arr[1] = '.';
	return 4;
}

//-------------------------------------------  
void displaySetRdsStation(uint8_t rdsReady, char* station){
	if (!rdsReady) {
		if (rdsStation[0]) updateFlags |= ufRdsStation;
		rdsStation[0] = 0;
	} else { // rds ready
		uint8_t i, j = 1;
		for (i = 0; i< RDS_NAME_LENGTH+1; i++ ) {
			if (station[i] != rdsStation[i]) j = 0;
			rdsStation[i] = station[i]; 
		}//for
		if (!j) updateFlags |= ufRdsStation;
	} // else
}

void displaySetSleepTime(uint8_t sleepMin){
	if (sleepTimeVal == sleepMin) return;
	sleepTimeVal = sleepMin;
	updateFlags |= ufTime;
}

void displaySetScanProgress(uint8_t scanned){
	scanChVal = scanned;
	updateFlags |= ufFrequency;
}

void displaySetScreenEditValue(uint16_t val){
	screenEditVal = val;
	updateFlags |= ufMenuChange;
}

void displayMenuIsChange(void){
	updateFlags |= ufMenuChange;
}

void displaySetChannel(uint8_t newChannel){
	if (channelVal == newChannel) return;
	channelVal = newChannel;
	updateFlags |= ufChannel;
}

void displaySetBattery(uint8_t isChaging, uint16_t voltage){
	if (chargeVal == isChaging && batteryVal == voltage) return;
	chargeVal = isChaging;
	batteryVal = voltage;
	updateFlags |= ufBattery;
}

void displaySetReceiveFlags(uint8_t newLevel, uint8_t newStereoFlag){
	if (levelVal == newLevel && newStereoFlag == stereoVal) return;
	levelVal = newLevel;
	stereoVal = newStereoFlag;
	updateFlags |= ufReceiverFlags;
}

void displaySetFrequency(uint16_t newFreq){
	if (freqVal == newFreq) return;
	freqVal = newFreq;
	updateFlags |= ufFrequency;
}

void displaySetTime(uint8_t newHour, uint8_t newMin){
	if (hourVal == newHour && minVal == newMin) return;
	hourVal = newHour;
	minVal = newMin;
	if (
		//oldMode ==  ||
		//oldMode ==  ||
		oldMode == mShowRadio ||
		oldMode == mShowTime ) updateFlags |= ufTime;
}

void displaySetVolumeVal(uint8_t newVolume){
	if (volumeVal == newVolume) return;
	volumeVal = newVolume;
	updateFlags |= ufVolume;
}

uint8_t drawProcessIsEnd(void){
	if (longProcessEnd) {
		longProcessEnd = 0;
		return 1;
	}
	return 0;
}

//------------------------------------------- 
static void drawMenuScrollBar(uint8_t pos, uint8_t count){
	#define TOP_BAR			18
	#define TOP_SCROLL		(18+1)
	#define BOTTOM_BAR		DIS_B
	#define BOTTOM_SCROLL	(DIS_B-2)
	#define LEFT_EDGE		0
	#define RIGHT_EDGE		4
	
	setDrawColor(0);
	setBackgroundColor(0);
	rectangle(LEFT_EDGE,TOP_BAR,RIGHT_EDGE,BOTTOM_BAR,fmFill);
	setDrawColor(1);
	line ((LEFT_EDGE+RIGHT_EDGE)/2,TOP_BAR,(LEFT_EDGE+RIGHT_EDGE)/2,BOTTOM_BAR);
	//    16,     
	uint16_t totalHeight = (uint16_t)16*(BOTTOM_SCROLL-TOP_SCROLL+1); //    
	uint16_t itemHeight = totalHeight / count; //      
	uint16_t itemTop = ((uint16_t)16*TOP_SCROLL)+(uint16_t)itemHeight*pos;
	if (itemHeight<32) itemHeight = 32;
	rectangle(LEFT_EDGE,(uint16_t)itemTop/16, RIGHT_EDGE,((uint16_t)itemTop+itemHeight)/16, fmFill);
	
	#undef TOP_BAR		
	#undef TOP_SCROLL	
	#undef BOTTOM_BAR	
	#undef BOTTOM_SCROLL
	#undef LEFT_EDGE
	#undef RIGHT_EDGE
} 

static void drawNextArrow(uint8_t topLine){
	#define A_WIDTH			5
	#define LEFT_EDGE		(DIS_R-A_WIDTH+1)
	#define RIGHT_EDGE		DIS_R
	#define V_SHIFT			7

	setDrawColor(0);
	setBackgroundColor(0);
	rectangle(LEFT_EDGE,topLine,RIGHT_EDGE,topLine+15-1,fmFill);
	setDrawColor(1);
	uint8_t i, j = topLine+V_SHIFT;
	for (i=0; i<A_WIDTH; i++) {
		line(RIGHT_EDGE-i, j, RIGHT_EDGE-i, j+i*2);
		j--;
	}
	
	#undef A_WIDTH
	#undef V_SHIFT
	#undef LEFT_EDGE
	#undef RIGHT_EDGE
}

static void drawBatterySymbol(void){
	uint16_t lowLevel = eeprom_read_word(&eeSettings.battery.lowVoltage);
	uint16_t quadStep = ((420-lowLevel)+2)/4;
	uint8_t i = 0;
	while ( batteryVal > lowLevel && i < 4) {
		lowLevel += quadStep;
		i++;
	} // while
	if (chargeVal) i = 5;
	uint16_t palette[2] = {0, 1};
	putFSpriteCustom(BAT_MAIN_LEFT, BAT_MAIN_TOP,	(spriteType*)pgm_read_word(&imgBatteryArr[ i ]), palette);
}

static void drawBigFrequency(char* buff){
	buff[ int2Frequency(freqVal, buff, 0) ] = 0;
	buff[3] = 0;
	setTextSymbolsInterval(FREQ_MAIN_FREQ_SPACING);
	putString(FREQ_MAIN_MHZ_LEFT, FREQ_MAIN_TOP, &buff[0], BIG_FONT);
	putString(FREQ_MAIN_KHZ_LEFT, FREQ_MAIN_TOP, &buff[4], BIG_FONT);
	setTextSymbolsInterval(1);
	uint16_t palette[2] = {0, 1};
	putFSpriteCustom(FREQ_MAIN_FREQ_DOT_LEFT, FREQ_MAIN_TOP, &imgFreqDot, palette);
	putFSpriteCustom(FREQ_MAIN_MHZ_SYMB_LEFT, FREQ_MAIN_MHZ_SYMB_TOP, &imgMHzSymbol, palette);
}
	  
static void clearAndOnDisplay(uint8_t turnOnOnly){
	if (!displayPower) {
		displayPower = 1;
		ssd1306on();
		ssd1306reInit();
		ssd1306setContrast(255);
	}
	if (turnOnOnly) return;
	ssd1306clearDisplay();
}
	  
void drawDisplay(modeType drawMode){
	if (oldMode == 0xFF) initRandSeed(batteryVal);
	static uint16_t phase = 0;
	static uint8_t phaseStep = 0;
	static uint8_t stepDelay = 0;
	uint8_t i,j;
	uint16_t freq;
	menuItem* tempMenu1; 
	menuItem* tempMenu2;
	uint16_t palette[2] = {0, 1};
	//   
	static uint16_t longProcess = 0;
	if (longProcess) {
		longProcess--;
		if (!longProcess) longProcessEnd = 1;
	}
	//
	#ifdef USE_CONTRAST_SHIFT
	// contrast
	if (!longProcess && displayPower && contrastCounter) {
		contrastCounter--;
		if (contrastCounter >= MAX_CONTRAST_TIME) ssd1306setContrast(255);
		if (contrastCounter < 255 && !(contrastCounter & 0x01) )	ssd1306setContrast(contrastCounter/2+128);
	} // if display power on
	#endif
	//
	//      
	if (oldMode == drawMode && updateFlags == 0 && longProcess == 0) return;
	uint8_t changeMode = oldMode != drawMode;
	oldMode = drawMode;
	if (changeMode) longProcess = 0;
	// 
	char buff[16]; 
	setDrawColor(1);
	setBackgroundColor(0);
	switch (drawMode) {
		case mRadioSleep:		//=============================    
		case mSleep:			//=============================   ( )
		case mErrNoFmSleep:		//=============================    " "
			if (changeMode) {
				ssd1306off();
				displayPower = 0;
			}
			break;
		case mErrNoFm:			//=============================  -   
			if (changeMode) clearAndOnDisplay(0);
			putFString(3,17, PSTR("RDA5807 Error!"), BASE_FONT);
			break;
		case mSetTime:			//=============================  
			if (changeMode) clearAndOnDisplay(0);
			//        .  -      
			putFString(0,0,selectedMenuItem->Text,BASE_FONT); 
			line(0,14,DIS_R,14);
			// 
			setDrawColor(1-subMode);
			rectangle(16-1,2*16, 16-1+ 2*(1+BASE_FONT_W), 2*16+15, fmFill);
			setBackgroundColor(1-subMode);
			setDrawColor(subMode);
			buff[0]	= '0' + (timeEditor.hour >> 4);
			buff[1] = '0' + (timeEditor.hour & 0x0F);
			buff[2] = 0;
			putString(16, 1+2*16, buff, BASE_FONT);
			// 
			setDrawColor(subMode);
			rectangle(16-1+3*(1+BASE_FONT_W),2*16, 16-1+ 2*(1+BASE_FONT_W)+3*(1+BASE_FONT_W), 2*16+15, fmFill);
			setBackgroundColor(subMode);
			setDrawColor(1-subMode);
			buff[0]	= '0' + (timeEditor.min >> 4);
			buff[1] = '0' + (timeEditor.min & 0x0F);
			buff[2] = 0;
			putString(16+3*(1+BASE_FONT_W), 1+2*16, buff, BASE_FONT);
			// 
			setDrawColor(1);
			setBackgroundColor(0);
			putFString(16+2*(1+BASE_FONT_W),1+2*16,PSTR(":"),BASE_FONT);
			break;
		case mPresetList:		//=============================   
			/*if (changeMode) */clearAndOnDisplay(0);
			putFString(0,0,PSTR("FM-"),BASE_FONT);
			line(0,14,DIS_R,14);
			if (subMode>0) {
				if (subMode<FM_PRESETS-1) j = subMode-1;
				else j = subMode - 2;
			} else j = 0;
			for ( i=1; i<4; i++) {
				if (j == subMode)	{
					setDrawColor(1);
					rectangle(8,i*16,DIS_W-8+1,i*16+15, fmFill);
					setDrawColor(0);
					setBackgroundColor(1);
				} else {
					setDrawColor(1);
					setBackgroundColor(0);
				}
				int2charArr(j+1,2,'0',buff);
				if (buff[0]=='0')  buff[0] = ' ';
				buff[2] = '.';
				buff[3] = 0;
				putString(9, i*16+1, buff, BASE_FONT);
				freq = eeprom_read_word(&eeSettings.radio.presets[j]);
				if (freq) {
					buff [ int2Frequency(freq, buff, 1) ] = 0;
					putString(9+4*(1+BASE_FONT_W), i*16+1, buff, BASE_FONT);
				} else putFString(9+4*(1+BASE_FONT_W),i*16+1,PSTR(" -"),BASE_FONT);
				drawNextArrow(i*16+1);
				j++;
			} // for
			drawMenuScrollBar( subMode, FM_PRESETS);
			break;
		case mAutoScan:			//=============================  
			if (changeMode) {
				clearAndOnDisplay(0);
				putFString(0,0,PSTR(""),BASE_FONT);
				line(0,14,DIS_R,14);
				//					  12345678901234
				putFString(0,33,PSTR(": 0"),BASE_FONT);
				putFString(0,49,PSTR("   [] - "),BASE_FONT);
			}
			buff[ int2Frequency(freqVal, buff, 1) ] = 0;
			putString(0+2*(1+BASE_FONT_W),17,buff, BASE_FONT);
			buff[ int2charArr(scanChVal, 1, '0', buff) ] = 0;
			putString(0+9*(1+BASE_FONT_W),33,buff, BASE_FONT);
			break;
		case mMainMenu:			//=============================  
			//  
			clearAndOnDisplay(0);
			//  .
			setDrawColor(1);
			setBackgroundColor(0);
			tempMenu1 = MENU_PARENT;
			if ((void*)tempMenu1 != &NULL_ENTRY) putFString(0,0,tempMenu1->Text,BASE_FONT);
			else if (isPresetSubMenu) {
					putFString(0,0,PSTR(" "),BASE_FONT);
					buff [ int2charArr(subMode+1, 1, '0', buff) ] = 0;
					putString(0+(1+BASE_FONT_W)*7, 0, buff, BASE_FONT);
				} else putFString(0,0,PSTR(""),BASE_FONT);
			line(0,14,DIS_R,14);
			//     
			tempMenu1 = MENU_PREVIOUS; //  
			// ,      ,   -   
			if (tempMenu1 == &NULL_ENTRY) tempMenu1 =  selectedMenuItem;
			//  ,   , ,  ,     -
			tempMenu2 = MENU_NEXT;
			if ((void*)tempMenu2 == &NULL_ENTRY) {
				// 
				tempMenu2 = (menuItem*)pgm_read_word(&tempMenu1->Previous);// -
				if ((void*)tempMenu2 != &NULL_ENTRY) tempMenu1 = tempMenu2;	//    - -  
			}
			i = 1;
			while (i < 4) {
				if (tempMenu1 == selectedMenuItem)	{
					setDrawColor(1);
					rectangle(8,i*16,DIS_W-8+1,i*16+15, fmFill);
					setDrawColor(0);
					setBackgroundColor(1);
				} else {
					setDrawColor(1);
					setBackgroundColor(0);
				}
				if ((void*)tempMenu1 != &NULL_ENTRY) {
					putFString(9,i*16+1,tempMenu1->Text,BASE_FONT);
					tempMenu2 = (menuItem*)pgm_read_word(&tempMenu1->Child);
					//     ,  
					if ((void*)tempMenu2 != &NULL_ENTRY) drawNextArrow(i*16+1);
					//    ,  ,    . 
					j = pgm_read_byte(&tempMenu1->Command);
					switch (j) {
						case mcmdGoPresets: 
							drawNextArrow(i*16+1); //  mmPresets    .
							break;
						case mcmdMonoStereo:
							if (eeprom_read_byte(&eeSettings.radio.stereo)) putFString(DIS_W-8-(1+BASE_FONT_W)*6,i*16+1,PSTR(""),BASE_FONT);
							else putFString(DIS_W-8-(1+BASE_FONT_W)*6,i*16+1,PSTR("  "),BASE_FONT);
							break;
						case mcmdLoudness:
							if (eeprom_read_byte(&eeSettings.radio.bass)) putFString(DIS_W-8-(1+BASE_FONT_W)*4,i*16+1,PSTR(" "),BASE_FONT);
							else putFString(DIS_W-8-(1+BASE_FONT_W)*4,i*16+1,PSTR(""),BASE_FONT);
							break;
						case mcmdRDSmode:
							if (eeprom_read_byte(&eeSettings.radio.showRds)) putFString(DIS_W-8-(1+BASE_FONT_W)*4,i*16+1,PSTR(" "),BASE_FONT);
							else putFString(DIS_W-8-(1+BASE_FONT_W)*4,i*16+1,PSTR(""),BASE_FONT);
							break;
						case mcmdChSaveDel:
							freq = eeprom_read_word(&eeSettings.radio.presets[subMode]);
							if (freq) putFString(9,i*16+1,PSTR(""),BASE_FONT);
							else putFString(9,i*16+1,PSTR(". "),BASE_FONT);
							break;
						case mcmdSetBkLight:
							buff[ int2charArr(screenEditVal, 2, '0', buff) ] = 0;
							if (buff[0]=='0') buff[0] = ' ';
							putString(9+BASE_FONT_W+1,i*16+1,buff,BASE_FONT);
							break;
						case mcmdSetBatVolt:
						case mcmdSetLowBat:
							buff[ makeBattery(screenEditVal, buff) ] = 0;
							putString(9+BASE_FONT_W+1,i*16+1,buff,BASE_FONT);
							break;
						case mcmdChSetFreq:
							buff[ int2Frequency(screenEditVal*5, buff, 1) ] = 0;
							putString(9+BASE_FONT_W+1,i*16+1,buff,BASE_FONT);
							break;
						case mcmdSetSleep:
							if (sleepTimeVal) {
								uint8_t n = int2charArr(sleepTimeVal, 2, '0', buff);
								if (buff[0] == '0' ) buff[0] = ' ';
								buff[n++] = ' ';
								buff[n++] = '';
								buff[n++] = '';
								buff[n++] = '';
								buff[n  ] = 0;
								putString(DIS_W-4-(1+BASE_FONT_W)*7, i*16+1, buff, BASE_FONT);
							} else putFString(DIS_W-4-(1+BASE_FONT_W)*7, i*16+1, PSTR("  "), BASE_FONT);
					} // switch
					//   
					tempMenu1 = (menuItem*)pgm_read_word(&tempMenu1->Next);
				}
				i++;
			}// while
			//        
			tempMenu1 = selectedMenuItem;
			//     
			while ((menuItem*)pgm_read_word(&tempMenu1->Previous) != &NULL_ENTRY) { 
				tempMenu1 = (menuItem*)pgm_read_word(&tempMenu1->Previous);
			}
			//    . 
			i = 1; //        - i = 1
			j = 0; //   ,   .
			while ((menuItem*)pgm_read_word(&tempMenu1->Next) != &NULL_ENTRY) {
				tempMenu1 = (menuItem*)pgm_read_word(&tempMenu1->Next);
				if (tempMenu1 == selectedMenuItem)  j =i;
				i++;
			}//while
			if (i>3) drawMenuScrollBar(j,i);  //      -  
			break;
		case mShowRadio:		//=============================   
			if (changeMode) {
				clearAndOnDisplay(0);
				updateFlags |= ufFrequency | ufReceiverFlags | ufVolume | ufTime | ufBattery | ufChannel | ufRdsStation;
			}
			// 
			if ( updateFlags & ufFrequency) {
				drawBigFrequency(buff);
			}
			//     
			if ( updateFlags & ufReceiverFlags ) {
				//  
				i = levelVal/16; // 
				putFSpriteCustom(FM_LEVEL_LEFT, FM_LEVEL_TOP,	(spriteType*)pgm_read_word(&imgFmLevels[ i ]), palette);
				// /
				if (stereoVal && eeprom_read_byte(&eeSettings.radio.stereo)) {
					putFSpriteCustom(STEREO_SYMB_LEFT, STEREO_SYMB_TOP, &imgStereo, palette);
				} else {
					setDrawColor(0);
					rectangle(STEREO_SYMB_LEFT, STEREO_SYMB_TOP,STEREO_SYMB_LEFT+STEREO_SYMB_WIDTH-1, STEREO_SYMB_TOP+STEREO_SYMB_HEIGHT-1, fmFill);
					setDrawColor(1);
				}
			}
			// 
			if (updateFlags & ufVolume) {
				putFSpriteCustom(MAIN_SCR_VOLUME_SYM_LEFT, MAIN_SCR_VOLUME_SYM_TOP, &imgVolumeSmall, palette);
				rectangle(MAIN_SCR_VOLUME_BAR_LEFT, MAIN_SCR_VOLUME_BAR_TOP, MAIN_SCR_VOLUME_BAR_LEFT+MAIN_SCR_VOLUME_BAR_WIDTH-1, MAIN_SCR_VOLUME_BAR_TOP+MAIN_SCR_VOLUME_BAR_HEIGHT-1, fmNoFill);
				setDrawColor(0);
				rectangle(MAIN_SCR_VOLUME_BAR_LEFT+1, MAIN_SCR_VOLUME_BAR_TOP+1, MAIN_SCR_VOLUME_BAR_LEFT+MAIN_SCR_VOLUME_BAR_WIDTH-1-1, MAIN_SCR_VOLUME_BAR_TOP+MAIN_SCR_VOLUME_BAR_HEIGHT-1-1, fmFill);
				setDrawColor(1);
				rectangle(MAIN_SCR_VOLUME_GAUGE_LEFT, MAIN_SCR_VOLUME_GAUGE_TOP, MAIN_SCR_VOLUME_GAUGE_LEFT+MAIN_SCR_VOLUME_GAUGE_STEP*volumeVal, MAIN_SCR_VOLUME_GAUGE_TOP+MAIN_SCR_VOLUME_GAUGE_HEIGHT-1, fmFill);
			}
			// 
			if (updateFlags & ufBattery) drawBatterySymbol();
			//
			//  
			if (updateFlags & ufChannel) {
				// 
				setDrawColor(0);
				rectangle(FREQ_MAIN_CHAN_AREA_LEFT, FREQ_MAIN_CHAN_AREA_TOP, FREQ_MAIN_CHAN_AREA_LEFT+FREQ_MAIN_CHAN_AREA_WIDTH-1, FREQ_MAIN_CHAN_AREA_TOP+FREQ_MAIN_CHAN_AREA_HEIGHT-1, fmFill);
				setDrawColor(1);
				if (channelVal < FM_PRESETS) {
					//    a - 
					rectangle(FREQ_MAIN_CHAN_AREA_LEFT, FREQ_MAIN_CHAN_AREA_TOP, FREQ_MAIN_CHAN_AREA_LEFT+FREQ_MAIN_CHAN_AREA_WIDTH-1, FREQ_MAIN_CHAN_AREA_TOP+FREQ_MAIN_CHAN_AREA_HEIGHT-1, fmNoFill);
					buff[ int2charArr(channelVal+1, 1, '0', buff) ] = 0;
					i = FREQ_MAIN_CHAN_VAL_LEFT;
					if (channelVal/*+1*/<10-1) i+= (1+SMALL_FONT_W)/2;
					putString(i, FREQ_MAIN_CHAN_VAL_TOP, buff, SMALL_FONT);
				}
			}
			//     
			if (updateFlags & ufTime) {
				setDrawColor(0);
				rectangle(MAIN_TIME_HOUR_LEFT, MAIN_TIME_TOP, MAIN_TIME_HOUR_LEFT+MAIN_TIME_WIDTH-1, MAIN_TIME_TOP+MAIN_TIME_HEIGHT-1, fmFill);
				setDrawColor(1);
				if (sleepTimeVal) {
					putFSpriteCustom(MAIN_TIME_ZZZ_LEFT, MAIN_TIME_TOP, &imgZzz, palette);
					buff[ int2charArr(sleepTimeVal,1,'0',buff) ] = 0;
					putString(MAIN_TIME_ZZZ_VAL_LEFT, MAIN_TIME_TOP, buff, BASE_FONT);
				} else {
					buff[ makeTime(buff) ] = 0;
					buff[2] = 0;
					putString(MAIN_TIME_HOUR_LEFT, MAIN_TIME_TOP, &buff[0], BASE_FONT);
					putString(MAIN_TIME_MINUTE_LEFT, MAIN_TIME_TOP, &buff[3], BASE_FONT);
					putFSpriteCustom(MAIN_TIME_DOTS_LEFT, MAIN_TIME_TOP, &imgTimeDoubleDot, palette);
				}
			}
			// RDS
			if (updateFlags & ufRdsStation) {
				setDrawColor(0);
				rectangle(RDS_STATION_AREA_LEFT, RDS_STATION_AREA_TOP, RDS_STATION_AREA_LEFT+RDS_STATION_AREA_WIDTH-1, RDS_STATION_AREA_TOP+RDS_STATION_AREA_HEIGHT-1, fmFill);
				setDrawColor(1);
				putString(RDS_STATION_AREA_LEFT, RDS_STATION_AREA_TOP, rdsStation, BASE_FONT);
			}
			break;
		case mRadioScanUp:		//=============================   /
		case mRadioScanDown:	//=============================   / 
			if (changeMode) {
				//clearAndOnDisplay(0);
				//clear channel icon & stereo icon
				setDrawColor(0);
				rectangle(FREQ_MAIN_CHAN_AREA_LEFT, FREQ_MAIN_CHAN_AREA_TOP, FREQ_MAIN_CHAN_AREA_LEFT+FREQ_MAIN_CHAN_AREA_WIDTH-1, FREQ_MAIN_CHAN_AREA_TOP+FREQ_MAIN_CHAN_AREA_HEIGHT-1, fmFill);
				rectangle(STEREO_SYMB_LEFT, STEREO_SYMB_TOP,STEREO_SYMB_LEFT+STEREO_SYMB_WIDTH-1, STEREO_SYMB_TOP+STEREO_SYMB_HEIGHT-1, fmFill);
				setDrawColor(1);
				// prepare animate
				phase = 0;
				stepDelay = 0;
				longProcess = 10000;
			}
			if (updateFlags & ufFrequency) drawBigFrequency(buff);
			if (stepDelay) stepDelay--;
			else {
				stepDelay = 80;
				switch (phase) {
					case 0:
						if (drawMode==mRadioScanUp)   putFString(RDS_STATION_AREA_LEFT,RDS_STATION_AREA_TOP, PSTR(">> >> >>"), BASE_FONT);
						if (drawMode==mRadioScanDown) putFString(RDS_STATION_AREA_LEFT,RDS_STATION_AREA_TOP, PSTR("<< << <<"), BASE_FONT);
						break;
					case 1:
						if (drawMode==mRadioScanUp)   putFString(RDS_STATION_AREA_LEFT,RDS_STATION_AREA_TOP, PSTR(" >> >> >"), BASE_FONT);
						if (drawMode==mRadioScanDown) putFString(RDS_STATION_AREA_LEFT,RDS_STATION_AREA_TOP, PSTR("< << << "), BASE_FONT);
						break;
					case 2:
						if (drawMode==mRadioScanUp)   putFString(RDS_STATION_AREA_LEFT,RDS_STATION_AREA_TOP, PSTR("> >> >> "), BASE_FONT);
						if (drawMode==mRadioScanDown) putFString(RDS_STATION_AREA_LEFT,RDS_STATION_AREA_TOP, PSTR(" << << <"), BASE_FONT);
						break;
				}  // switch
				phase++;
				if (phase>2) phase = 0;
			}
			break;
		case mShowTime:			//=============================   ( )
			if (changeMode) clearAndOnDisplay(0);
			drawBatterySymbol();
			i = makeBattery(batteryVal, buff);
			//buff[i++] = ' ';
			buff[i++] = '';
			buff[i] = 0;
			putString(DIS_W-(1+BASE_FONT_W)*i,1,buff,BASE_FONT);
			i = makeTime(buff);
			buff[i] = 0;
			setTextSymbolsInterval(2);
			putString(10,24, buff, BIG_FONT);
			setTextSymbolsInterval(1);
			break;
		case mRadioTurnOff:		//=============================  
			if (changeMode) {
				clearAndOnDisplay(1);
				phase = 4;  // 
				phaseStep = 0;
				longProcess = 10000;
			}// if chhangeMode
			if (!phaseStep && phase) {
				phase--;
				switch (phase) {
					case 0: 
						longProcess = 1;
						break;
					case 1:
						phaseStep = 4;
						break;
					case 2: 
						phaseStep = 61; 
						break; 
					case 3: 
						phaseStep = 32; 
						break; 
				} // switch
			} //if (!phaseStep && phase)
			if (phaseStep) {
				if (stepDelay ) stepDelay--;
				else {
					phaseStep--;
					switch(phase) {
						case 3:
							stepDelay = 0 + (32-phaseStep)/2;
							if (phaseStep<31) {
								setDrawColor(0);
								line(0,30-phaseStep,DIS_R,30-phaseStep);
								line(0,33+phaseStep,DIS_R,33+phaseStep);
							}
							setDrawColor(1);
							line(0,31-phaseStep,DIS_R,31-phaseStep);
							line(0,32+phaseStep,DIS_R,32+phaseStep);
							break;
						case 2:
							stepDelay = 0 + (63-phaseStep)/8;
							setDrawColor(0);
							line(60-phaseStep,31,60-phaseStep,32);	
							line(67+phaseStep,31,67+phaseStep,32);
							if (!phaseStep) {
								//  
								setDrawColor(1);
								rectangle(63,29,64,34,fmNoFill);
								rectangle(62,30,65,33,fmNoFill);
								rectangle(61,31,66,32,fmNoFill);
							}
							break;
						case 1: 
							stepDelay = 200;
							setDrawColor(0);
							switch(phaseStep) {
								case 3:
									rectangle(61,29,66,34,fmNoFill);
									break;
								case 2:
									displayPutPixel(62,30,0);
									displayPutPixel(65,30,0);
									displayPutPixel(62,33,0);
									displayPutPixel(65,33,0);
									break;
								case 1:
									rectangle(62,30,65,33,fmNoFill);
									break;
							}
							//displayPutPixel(x,y,c);
							break;
					}	// switch
				}	// else (id StepDelay)
			} // if (phaseStep)
			break;
		case mBatteryLow:		//=============================   
		case mNoHeadphones:		//=============================  
			if (changeMode) {
				clearAndOnDisplay(0);
				longProcess = 4098;	 						//		1234567890123456
				if (drawMode == mNoHeadphones) putFString(0,0,PSTR("  !"),BASE_FONT);
				if (drawMode == mBatteryLow)   putFString(0,0,PSTR(". !"),BASE_FONT);
			}
			if ( !(longProcess & 0x00FF)) {
				i = 0;
				if (longProcess & 0x0100) i = 1;
				if (drawMode == mNoHeadphones) putFSpriteCustom(NO_PHONES_LEFT, NO_PHONES_TOP,	(spriteType*)pgm_read_word(&imgNoPhones[ i ]),  palette);
				if (drawMode == mBatteryLow) putFSpriteCustom(LOW_BATTERY_LEFT, LOW_BATTERY_TOP,(spriteType*)pgm_read_word(&imgLoBattery[ i ]), palette);
			}
			break;
		case mSplashScreen:		//=============================     
		case mRadioTurnOn:		//=============================    
			if (changeMode) {
				clearAndOnDisplay(0);
				phase = 0;  // 
				phaseStep = 2;	  // 2 -   , 1 -  
				if (drawMode == mRadioTurnOn && getRand(16)) phaseStep = 1; 
				stepDelay = 0;
				longProcess = 1000;
			}// if chhangeMode
			if (phaseStep) {
				if (stepDelay) stepDelay--;
				else {
					if (drawMode == mSplashScreen) stepDelay = 50;
					else stepDelay = 5;
					if (phaseStep == 2) {
						i = pgm_read_byte(&splashScreen1[phase].cX);
						j = pgm_read_byte(&splashScreen1[phase].cY);
					} else {
						i = pgm_read_byte(&splashScreen2[phase].cX);
						j = pgm_read_byte(&splashScreen2[phase].cY);
					}
					phase++;
					if (drawMode == mSplashScreen) longProcess = 2000;
					else longProcess = 500;
					if (j & SSY_END_DRAW) {
						phaseStep--;
						phase = 0;
					}
					uint8_t newl = i & SSX_START_LINE;
					i &= ~SSX_START_LINE;
					j &= ~SSY_END_DRAW;
					if (newl) line(i,j,i,j);
					else lineTo(i,j);
				} // else
			} // if (phaseStep)
			break;				
		case mRadioGoSleep:		//=============================      
		case mTimeGoSleep:		//=============================      
			//   
			if (changeMode) {
				clearAndOnDisplay(1);
				phase = 128;  // 
				stepDelay = 0;
				longProcess = 10000;
			}// if chhangeMode
			if (phase) {
				if (stepDelay) stepDelay--;
				else {
					//stepDelay = 2;
					phase--;
					if (!phase) longProcess = 1;
					i = phase / 2; // i = 63-->0
					setDrawColor(0);
					if (phase & 0x01) line(63-i,0, 63-i,DIS_B);
					else line(64+i,0, 64+i,DIS_B);
				} // else
			} // if (phaseStep)	 //*/
			break;				

	} // switch (mode)
	if (displayPower) ssd1306updateDisplay();
	updateFlags = 0;
}

#undef MAX_CONTRAST_TIME