/*
 * FM_RCVR_SSD1306_RDA5807.c
 *
 * Created: 23.12.2019 13:58:09
 * Author : GoldenAndy
 */

#include "defines.h"

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <string.h>
#include <avr/pgmspace.h>

#include "main.h"
#include "HAL.h"
#include "graphics.h"
#include "menu.h"
#include "../_common/keyAntiBounce/keyAntiBounce.h"
#include "../_common/systemClock/systemClock.h"
#include "../_common/EEPROM_CheckSum/EEPROM_CheckSum.h"
#include "../_common/RDA5807M/RDA5807M.h"

typedef enum {
	noPowerAction,
	setRadioPowerOff,
	setRadioPowerOn
	} radioPowerControlType;

//-------------------------------------------  

modeType mode = mSplashScreen;
volatile uint8_t soundPlayCounter = 0;
uint16_t modeTimeOut = 0;
keyProcessingStruct keyUp, keyDn, keyLeft, keyRight, keyOk;
volatile timeType time = { .hour = 0, .min = 0, .sec = 0, .changed = 1 };
volatile uint8_t keyWake = 0;
uint8_t subMode = 0;
timeEditorType timeEditor = {0,0};
uint8_t isPresetSubMenu = 0;

EEMEM settingsType eeSettings;

//------------------------------------------- 
uint8_t incBcd(uint8_t bcd);
uint8_t decBcd(uint8_t bcd);
 
// -------------------------------------------  
//MAKE_MENU (mmEntryName,		NULL_ENTRY,		NULL_ENTRY,		NULL_ENTRY,		NULL_ENTRY,		0,				"Entry text");
// 	      (Name,			Next,			Previous,		Parent,			Child,			Command,		 12345678901234  Text) 
// level 1
MAKE_MENU (mmPresets,		mmSleep,		NULL_ENTRY,		NULL_ENTRY,		NULL_ENTRY,		mcmdGoPresets,	"FM-");
MAKE_MENU (mmSleep,			mmRadioSetup,	mmPresets,		NULL_ENTRY,		NULL_ENTRY,		mcmdSetSleep,	"");
MAKE_MENU (mmRadioSetup,	mmDevSetup,		mmSleep,		NULL_ENTRY,		mmMonoStereo,	0,				".");
MAKE_MENU (mmDevSetup,		NULL_ENTRY,		mmRadioSetup,	NULL_ENTRY,		mmTimeSetup,	0,				".");
// special level - channel's manage																			 12345678901234
MAKE_MENU (mmChSaveDel,		mmChMoveUp,		NULL_ENTRY,		NULL_ENTRY,		NULL_ENTRY,		mcmdChSaveDel,	""/*/*/);//     
MAKE_MENU (mmChMoveUp,		mmChMoveDn,		mmChSaveDel,	NULL_ENTRY,		NULL_ENTRY,		mcmdMoveChUp,	" ");
MAKE_MENU (mmChMoveDn,		mmChInsLine,	mmChMoveUp,		NULL_ENTRY,		NULL_ENTRY,		mcmdMoveChDn,	" ");
MAKE_MENU (mmChInsLine,		mmChDelLine,	mmChMoveDn,		NULL_ENTRY,		NULL_ENTRY,		mcmdChInsLine,	". ");
MAKE_MENU (mmChDelLine,		mmChSetFreq,	mmChInsLine,	NULL_ENTRY,		NULL_ENTRY,		mcmdChDelLine,	" ");
MAKE_MENU (mmChSetFreq,		NULL_ENTRY,		mmChDelLine,	NULL_ENTRY,		mmChEditFreq,	0,				" ");
// special level+1 - channel's manage-  														 12345678901234
MAKE_MENU (mmChEditFreq,	NULL_ENTRY,		NULL_ENTRY,		mmChSetFreq,	NULL_ENTRY,		mcmdChSetFreq,	"");
//    mmRadioSetup																				 12345678901234
MAKE_MENU (mmMonoStereo,	mmLoudness,		NULL_ENTRY,		mmRadioSetup,	NULL_ENTRY,		mcmdMonoStereo,	":");
MAKE_MENU (mmLoudness,		mmRDSmode,		mmMonoStereo,	mmRadioSetup,	NULL_ENTRY,		mcmdLoudness,	".:");
MAKE_MENU (mmRDSmode,		mmAutoScan,		mmLoudness,		mmRadioSetup,	NULL_ENTRY,		mcmdRDSmode,	" RDS:");
MAKE_MENU (mmAutoScan,		NULL_ENTRY,		mmRDSmode,		mmRadioSetup,	mmAScanCancel,	0,				"");
//  3  																					 12345678901234
MAKE_MENU (mmAScanCancel,	mmAScanStart,	NULL_ENTRY,		mmAutoScan,		NULL_ENTRY,		0,				"");
MAKE_MENU (mmAScanStart,	NULL_ENTRY,		mmAScanCancel,	mmAutoScan,		NULL_ENTRY,		mcmdAutoScan,	"");
// 	      (Name,			Next,			Previous,		Parent,			Child,			Command,		 12345678901234  Text)
//  2   mmDevSetup																				 12345678901234
MAKE_MENU (mmTimeSetup,		mmBkLightTime,	NULL_ENTRY,		mmDevSetup,		mmTimeSetupEd,	0,				".");
MAKE_MENU (mmBkLightTime,	mmKeySound,		mmTimeSetup,	mmDevSetup,		mmSetBkLight,	0,				"...");
MAKE_MENU (mmKeySound,		mmKeyAction,	mmBkLightTime,	mmDevSetup,		mmSetKeySound1,	0,				" ");
MAKE_MENU (mmKeyAction,		mmBatteryCalib,	mmKeySound,		mmDevSetup,		mmKeyActOff,	0,				". ..");
MAKE_MENU (mmBatteryCalib,	mmLowBatLevel,	mmKeyAction,	mmDevSetup,		mmSetBatVolt,	0,				".");
MAKE_MENU (mmLowBatLevel,	mmDevRestart,	mmBatteryCalib,	mmDevSetup,		mmSetLowLevel,	0,				"...");
MAKE_MENU (mmDevRestart,	NULL_ENTRY,		mmLowBatLevel,	mmDevSetup,		mmDevRestart1,	0,				" ");
//  3   mmTimeSetup																				 12345678901234
MAKE_MENU (mmTimeSetupEd,	NULL_ENTRY,		NULL_ENTRY,		mmTimeSetup,	NULL_ENTRY,		mcmdSetTime,	".");
//  3  mmBkLightTime																				 12345678901234
MAKE_MENU (mmSetBkLight,	NULL_ENTRY,		NULL_ENTRY,		mmBkLightTime,	NULL_ENTRY,		mcmdSetBkLight,	" __ .");
//  3  mmKeySound																				 12345678901234
MAKE_MENU (mmSetKeySound1,	mmSetKeySound0,	NULL_ENTRY,		mmKeySound,		NULL_ENTRY,		mcmdKeySound1,	"");
MAKE_MENU (mmSetKeySound0,	NULL_ENTRY,		mmSetKeySound1,	mmKeySound,		NULL_ENTRY,		mcmdKeySound0,	"");
// 	      (Name,			Next,			Previous,		Parent,			Child,			Command,		 12345678901234  Text)
//  3   mmKeyAction																				 12345678901234
MAKE_MENU (mmKeyActOff,		mmKeyActOn,		NULL_ENTRY,		mmKeyAction,	NULL_ENTRY,		mcmdKeyActOff,	" .");
MAKE_MENU (mmKeyActOn,		NULL_ENTRY,		mmKeyActOff,	mmKeyAction,	NULL_ENTRY,		mcmdKeyActOn,	". .");
//  3   mmBatteryCalib																			 12345678901234
MAKE_MENU (mmSetBatVolt,	NULL_ENTRY,		NULL_ENTRY,		mmBatteryCalib,	NULL_ENTRY,		mcmdSetBatVolt,	" _.__ ");
//  3   mmLowBatLevel																				 12345678901234
MAKE_MENU (mmSetLowLevel,	NULL_ENTRY,		NULL_ENTRY,		mmLowBatLevel,	NULL_ENTRY,		mcmdSetLowBat,	" _.__ ");
//  3   mmDevRestart																				 12345678901234
MAKE_MENU (mmDevRestart1,	mmDevRestart2,	NULL_ENTRY,		mmDevRestart,	NULL_ENTRY,		0,				"");
MAKE_MENU (mmDevRestart2,	NULL_ENTRY,		mmDevRestart1,	mmDevRestart,	mmDevRestart3,	0,				" ");
//  4   mmDevRestart																				 12345678901234
MAKE_MENU (mmDevRestart3,	mmDevRestart4,	NULL_ENTRY,		mmDevRestart,	NULL_ENTRY,		0,				"");
MAKE_MENU (mmDevRestart4,	mmDevRestart5,	mmDevRestart3,	mmDevRestart,	NULL_ENTRY,		0,				"");
MAKE_MENU (mmDevRestart5,	NULL_ENTRY,		mmDevRestart4,	mmDevRestart,	NULL_ENTRY,		mcmdDevClear,	" ");

//MAKE_MENU (mmEntryName,		NULL_ENTRY,		NULL_ENTRY,		NULL_ENTRY,		NULL_ENTRY,		0,				"Entry text");
 
//-------------------------------------------    
ISR(ONE_SECOND_INTERRUPT){
	time.changed = 1;
	//
	time.sec = incBcd(time.sec);
	if (time.sec == 0x60) {
		time.sec = 0;
		time.min = incBcd(time.min);
		if (time.min == 0x60) {
			time.min = 0;
			time.hour = incBcd(time.hour);
			if (time.hour == 0x24) time.hour = 0;
		} // min == 60
	} // sec == 60
}

//-------------------------------------------   
ISR(KEYS_WAKE_INTERRUPT){
	keyWake = 1; 
}

//-------------------------------------------    
ISR(SYS_CLOCK_INTERRUPT){
#define SUB_TOP	4
	//
	//  
	static uint8_t soundFlag = 0;
	static uint8_t soundVolume = 0;
	if (soundPlayCounter) {
		//   
		if (!soundFlag) {
			//     
			soundVolume = eeprom_read_byte(&eeSettings.system.keySound);
			//    
			if (soundVolume) phonesEnableOutput();
			soundFlag = 1;
		}
		if (soundVolume) phonesSetOut(soundFlag & 2);
		soundFlag ^= 2;
	} else { // if (soundPlayCounter)
		if (soundFlag) {
			if (soundVolume) phonesEnableCheckPresent();
			soundFlag = 0;
		}
	} // else .. if (soundPlayCounter)
	//
	// sys clock
	static uint8_t subCounter = 0;
	if (subCounter < SUB_TOP-1) subCounter++;
	else {
		subCounter = 0;
		systemClockTick();
		if (soundPlayCounter) soundPlayCounter--;
	}
#undef SUB_TOP
} 

//------------------------------------------- BCD routines 
uint8_t incBcd(uint8_t bcd){
	if ((bcd & 0x0F) > 8) return bcd + 7;
	return bcd+1;
}
//----
uint8_t decBcd(uint8_t bcd){
	if (bcd & 0x0F) return bcd-1;
	return bcd-7;
}

//-------------------------------------------    
uint8_t findNearChannel(uint16_t freqSearch, uint8_t exact){
	// freqSearch - ,      
	// exactOnly  -    
	//       (exact == 0) -   , 
	//     (  )     -  0xFF
	uint8_t result = 0xFF;
	uint16_t freq;
	uint16_t delta = 0xFFFF;
	uint8_t i = 0;
	while (i<FM_PRESETS) {
		freq = eeprom_read_word(&eeSettings.radio.presets[i]);
		if (freq) {
			int16_t df = freq - freqSearch;
			if (df<0) df = -df;
			if (df == 0 ) return i;
			if (delta > df) {
				delta = df;
				result = i;
			}
		}// if freq
		i++;
	}	// while i
	if (delta && exact) return 0xFF;
	return result;
}

uint8_t getNextChannel (uint8_t channel, uint8_t findUp){
	if (channel == 0xFF) return channel;
	uint8_t i = channel;
	do {
		if (findUp) {
			if (i<FM_PRESETS-1) i++; 
			else i = 0;
		} else {
			if (i>0) i--;
			else i = FM_PRESETS-1;
		}
		uint16_t freq = eeprom_read_word(&eeSettings.radio.presets[i]);
		if (freq) return i;
	} while (i != channel);
	return channel;
}

//-------------------------------------------  
void init(void){
	//  ,      
	WDR;
	MCUSR &= ~(1<<WDRF);/* Clear WDRF in MCUSR */
	WDTCSR |= (1<<WDCE) | (1<<WDE);
	WDTCSR = 0x00;	/* Turn off WDT */
	//
	_delay_ms(500);
	hardwareInit();
	// 
	keyProcessingInit(&keyOk,    _KEY_FLAG_SHORTPRESS_ | _KEY_FLAG_LONGPRESS_);
	keyProcessingInit(&keyLeft,  _KEY_FLAG_SHORTPRESS_ | _KEY_FLAG_LONGPRESS_);
	keyProcessingInit(&keyRight, _KEY_FLAG_SHORTPRESS_ | _KEY_FLAG_LONGPRESS_);
	keyProcessingInit(&keyUp,    _KEY_FLAG_SHORTPRESS_ | _KEY_FLAG_REPEAT_);
	keyProcessingInit(&keyDn,    _KEY_FLAG_SHORTPRESS_ | _KEY_FLAG_REPEAT_);
	// display init
	ssd1306setContrast(255);
	ssd1306setOrientation(1,1);
	ssd1306setInversion(0);
	// display on
	ssd1306on();
	// display clear
	ssd1306clearDisplay();
	ssd1306updateDisplay();
	// FM init
	rda5807mErrorType rdaerr =  rda5807mInit();
	if (rdaerr != rda_Ok) {
		mode = mErrNoFm;
		modeTimeOut = 5000;
	} else {
		mode = mSplashScreen;
		modeTimeOut = 0;
	}	 
	uint8_t i;
	//   eeprom
	uint8_t eeVer = eeprom_read_byte(&eeSettings.eeVersion);
	if (!EEPROM_IsGoodCRC() || eeVer != EEVERSION) {
		eeprom_update_byte(&eeSettings.eeVersion, EEVERSION);
		for (i=0; i < FM_PRESETS; i++) eeprom_update_word(&eeSettings.radio.presets[i], 0);
		eeprom_update_word(&eeSettings.radio.lastFreq, 8700);
		eeprom_update_byte(&eeSettings.radio.lastVol, 4);
		eeprom_update_byte(&eeSettings.radio.bass, 1);
		eeprom_update_byte(&eeSettings.radio.stereo, 1);
		eeprom_update_byte(&eeSettings.radio.showRds, 1);
		eeprom_update_byte(&eeSettings.system.lightTime, 8);
		eeprom_update_byte(&eeSettings.system.keySound, 1);
		eeprom_update_byte(&eeSettings.system.keyAction, 1);
		eeprom_update_byte(&eeSettings.battery.bandGap, 110);
		eeprom_update_word(&eeSettings.battery.lowVoltage, 330);
		EEPROM_UpdateCRC();
	}
	//  
	sei();
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

int main(void)
{
	init();

	uint8_t multiPressFlag = 0;
	uint8_t resetKeysBySleep = 0;
	uint8_t radioVolume = 0;
	uint8_t radioPower = 0;
	uint8_t currentChannel = 0xFF;
	uint8_t isScreenEdit = 0;
	uint8_t rollingValue = 0;
	uint16_t onScreenEditVal=0, onScreenEditMin=0, onScreenEditMax=0;
	uint8_t updateBatteryFilter = 1;
	uint16_t batFilter[BATTERY_FILTER_DEPTH];
	uint8_t batFilterPtr = 0;
	uint16_t batVoltage = 400;
	uint8_t keyWatchCounter = 0;
	uint16_t saveFreq = 0;
	uint8_t scanChannel = 0;
	uint8_t stopScan = 0;
	uint8_t sleepMin = 0;
	uint8_t sleepSec = 0;
	
    while (1) {	// SUPER LOOP!
		//-------------------------------------------  ,    
		resetKeysBySleep = 0;
		modeType newMode = mode;
		uint8_t reloadTimeOut = 0;
		uint8_t i; //   
		uint8_t keyRepeat = 0;
		systemClockCounterType clk;
		radioPowerControlType radioPowerControl = noPowerAction;

		//-------------------------------------------   
		do {
			clk = getSysClockInterval();
		} while (!clk);

		//-------------------------------------------  ,    
		uint8_t draftKeys = keysGetPress();
		//      ,   
		if ( ( (mode == mSleep || mode == mErrNoFmSleep) && draftKeys & KEY_CENTER)  
			||  (mode == mRadioSleep && draftKeys) ) keyWatchCounter = 250;
		// ,       
		uint8_t multiPress = 0;
		if (draftKeys & KEY_CENTER) multiPress++;
		if (draftKeys & KEY_DOWN)   multiPress++;
		if (draftKeys & KEY_UP)     multiPress++;
		if (draftKeys & KEY_LEFT)   multiPress++;
		if (draftKeys & KEY_RIGHT)  multiPress++;
		//     -    multiPress	-     
		if (multiPress == 0) multiPressFlag = 0;
		if (multiPress == 1) multiPress = 0;	//     
		if (multiPress >  0) multiPressFlag = 1;
		//     
		keyPressType keyPress = kpNoPress;
		// key OK
		switch (keyProcessing(&keyOk, draftKeys & KEY_CENTER, multiPressFlag | resetKeysBySleep)) {
			case _KEY_EVENT_SHORTPRESS_: 
				keyPress = kpOk;
				break;
			case _KEY_EVENT_LONGPRESS_:
				keyPress = kpOkLong;
				break;
		} // switch
		// key LEFT
		switch (keyProcessing(&keyLeft, draftKeys & KEY_LEFT, multiPressFlag | resetKeysBySleep)) {
			case _KEY_EVENT_SHORTPRESS_: 
				keyPress = kpLeft;
				break;
			case _KEY_EVENT_LONGPRESS_:
				keyPress = kpLeftLong;
				break;
		} // switch
		// key RIGHT
		switch (keyProcessing(&keyRight, draftKeys & KEY_RIGHT, multiPressFlag | resetKeysBySleep)) {
			case _KEY_EVENT_SHORTPRESS_: 
				keyPress = kpRight;
				break;
			case _KEY_EVENT_LONGPRESS_:
				keyPress = kpRightLong;
				break;
		} // switch
		uint8_t keyFlags = 0;
		// key UP
		keyFlags = keyProcessing(&keyUp, draftKeys & KEY_UP, multiPressFlag | resetKeysBySleep);
		if (keyFlags) {
			keyPress = kpUp;
			keyRepeat |=  keyFlags & _KEY_EVENT_REPEAT_;
		}
		// key DOWN
		keyFlags = keyProcessing(&keyDn, draftKeys & KEY_DOWN, multiPressFlag | resetKeysBySleep);
		if (keyFlags) {
			keyPress = kpDown;
			keyRepeat |=  keyFlags & _KEY_EVENT_REPEAT_;
		}
		
		//-------------------------------------------   
		if (keyPress) {
			#ifdef USE_CONTRAST_SHIFT
			displaySetMaxContrast();
			#endif
			reloadTimeOut = 1;	 //    
			//  
			if (mode != mSleep && mode != mShowTime && mode != mTimeGoSleep && !keyRepeat) soundPlayCounter = KEY_SOUND_DURATION;
		}
		
		//-------------------------------------------  
		if (time.changed) {
			time.changed = 0;
			//   
			displaySetTime(time.hour, time.min);
			//  
			uint8_t bandGap = eeprom_read_byte(&eeSettings.battery.bandGap);
			//      bandgap-     
			if (updateBatteryFilter) {
				for (i=0; i < BATTERY_FILTER_DEPTH; i++) batFilter[i] = batteryGetVoltage(bandGap);
				updateBatteryFilter = 0;
			} // if updateBatteryFilter 
			//   -     .
			batFilter[batFilterPtr] = batteryGetVoltage(bandGap);
			batFilterPtr++;
			if (batFilterPtr == BATTERY_FILTER_DEPTH) batFilterPtr = 0;
			batVoltage = BATTERY_FILTER_DEPTH/2;
			for (i=0; i<BATTERY_FILTER_DEPTH; i++) batVoltage += batFilter[i];
			batVoltage /= BATTERY_FILTER_DEPTH;
			displaySetBattery(batteryGetChargingState(), batVoltage);
			//  
			if (( mode == mShowRadio || mode == mRadioSleep ) && sleepMin ) {
				if (sleepSec) sleepSec--;
				else {
					sleepSec = 59;
					if (sleepMin) {
						sleepMin--;
						displaySetSleepTime(sleepMin);
						if (!sleepMin) {
							radioPowerControl = setRadioPowerOff;
							newMode = mRadioTurnOff;
						}  //  if !sleepMin
					} // if sleepMin
				} // if sleepsec ... else
				if (sleepMin == 1 && (sleepSec>54 || (sleepSec<25 && sleepSec>=20) ) ) soundPlayCounter = KEY_SOUND_DURATION;
			}
			//    3   
			if (radioPower) {
				if (batVoltage <= eeprom_read_word(&eeSettings.battery.lowVoltage) ) {
					radioPowerControl = setRadioPowerOff;
					newMode = mBatteryLow;
				}
			}
			//   
			if ( radioPower && (mode == mShowRadio || mode == mRadioSleep) && !phonesGetPresent())	{
				radioPowerControl = setRadioPowerOff;
				newMode = mNoHeadphones;
			}
		}
		
		
		//-------------------------------------------  
		if (keyPress != kpNoPress && mode == newMode) { //    -    
			switch (mode) {
				case mSplashScreen:		//=============================     
				case mRadioTurnOn:		//=============================    
				case mRadioScanUp:		//=============================   /
				case mRadioScanDown:	//=============================   /
				case mRadioTurnOff:		//=============================  
				case mNoHeadphones:		//=============================  
				case mErrNoFm:			//=============================  -   
				case mBatteryLow:		//=============================   
					break;
				case mSleep:			//=============================   ( )
					switch(keyPress) {
						case kpOk:			//------------ key OK
							newMode = mShowTime;
							break;
						case kpOkLong:		//------------ key OK (long press)
							newMode = mRadioTurnOn;
							break;
						default:
							break;
					} // switch (keyPress)
					break;				
				case mShowTime:			//=============================   ( )
				case mTimeGoSleep:		//=============================      
					if (keyPress == kpOk) modeTimeOut = 1;
					if (keyPress == kpOkLong) newMode = mRadioTurnOn;
					break;				
				case mRadioSleep:		//=============================    
				case mRadioGoSleep:		//=============================      
					newMode = mShowRadio;
					// break,       
					if ( !eeprom_read_byte(&eeSettings.system.keyAction)) break; //    
					//     -    - 
					//         
				case mShowRadio:		//=============================   
					switch(keyPress) {
						case kpOk:			//------------ key OK
							menuChange(&mmPresets);
							newMode = mMainMenu;
							break;
						case kpOkLong:		//------------ key OK (long press)
							radioPowerControl = setRadioPowerOff;
							newMode = mRadioTurnOff;
							break;
						case kpLeft:		//------------ key LEFT
						case kpRight:		//------------ key RIGHT
							if (currentChannel == 0xFF) {
								rda5807mGetErrorState(1);
								i = findNearChannel(rda5807mGetFrequency(), 0);
								if (i == 0xFF) {
									//    -  .   ....
									if (keyPress == kpRight) keyPress = kpRightLong;
									else keyPress = kpLeftLong;
									goto tryScanNextStation;
								}
							} else {
								i = getNextChannel(currentChannel, keyPress == kpRight);
							}
							currentChannel = i;
							if (currentChannel != 0xFF)  {
								uint16_t freq = eeprom_read_word(&eeSettings.radio.presets[currentChannel]);
								if (freq) {	
									rda5807mSetFrequency(freq);
									rda5807mGetErrorState(1);
								}
								displaySetChannel(currentChannel);
							}
							break;
						case kpLeftLong:	//------------ key LEFT (long press)
						case kpRightLong:	//------------ key RIGHT (long press)
	tryScanNextStation:
							// scan
							rda5807mGetErrorState(1);
							if (keyPress == kpLeftLong) {
								rda5807mSearch(tuneDown);
								newMode = mRadioScanDown;
							}
							if (keyPress == kpRightLong) {
								rda5807mSearch(tuneUp);
								newMode = mRadioScanUp;
							}
							break;
						case kpUp:			//------------ key UP
						case kpDown:		//------------ key DOWN
							if (keyPress == kpUp && radioVolume<16) radioVolume++;
							if (keyPress == kpDown && radioVolume)  radioVolume--;
							if (radioVolume) {
								rda5807mSetMuteMode(muteOff);
								rda5807mSetVolume(radioVolume-1);
							} else {
								rda5807mSetMuteMode(muteOn);
								rda5807mSetVolume(0);
							}
							displaySetVolumeVal(radioVolume);
							break;
					} // switch (keyPress)
					break;				
				case mErrNoFmSleep:	    //=============================    " "
					switch(keyPress) {
						case kpOk:			//------------ key OK
						case kpOkLong:		//------------ key OK (long press)
							newMode = mErrNoFm;
							break;
						default:
							break;
					} // switch (keyPress)
					break;
				case mSetTime:			//  
					switch(keyPress) {
						case kpOk:			//------------ key OK
							cli();
							time.hour = timeEditor.hour;
							time.min = timeEditor.min;
							time.sec = 0;
							sei();
							newMode = mMainMenu;
							menuChange(&mmTimeSetup);
							break;
						case kpOkLong:		//------------ key OK (long press)
							radioPowerControl = setRadioPowerOff;
							newMode = mRadioTurnOff;
							break;
						case kpLeftLong:	//------------ key LEFT (long press)
							isPresetSubMenu = 0;
							newMode = mShowRadio;
							currentChannel = findNearChannel(rda5807mGetFrequency(), 1);
							displaySetChannel(currentChannel);
							break;
						case kpLeft:		//------------ key LEFT
							if (subMode) {
								subMode--;
								displayMenuIsChange();
							} else {
								newMode = mMainMenu;
								menuChange(&mmTimeSetup);
							}
							break;
						case kpRight:		//------------ key RIGHT
						case kpRightLong:	//------------ key RIGHT (long press)
							if (!subMode) {
								subMode++;
								displayMenuIsChange();
								}
							break;
						case kpUp:			//------------ key UP
							if (subMode == 0) {
								if (timeEditor.hour<0x23) timeEditor.hour = incBcd(timeEditor.hour);
								else timeEditor.hour = 0x00;
							} else {
								if (timeEditor.min<0x59) timeEditor.min = incBcd(timeEditor.min);
								else timeEditor.min = 0x00;
							}
							displayMenuIsChange();
							break;
						case kpDown:		//------------ key DOWN
							if (subMode == 0) {
								if (timeEditor.hour>0x00) timeEditor.hour = decBcd(timeEditor.hour);
								else timeEditor.hour = 0x23;
							} else {
								if (timeEditor.min>0x00) timeEditor.min = decBcd(timeEditor.min);
								else timeEditor.min = 0x59;
							}
							displayMenuIsChange();
							break;
					} // switch (keyPress)	
					break;
				case mAutoScan:			//=============================  
					switch(keyPress) {
						case kpOk:			//------------ key OK
							stopScan = 1;
							break;
						default:	
							break;
					} // switch (keyPress)	
				case mPresetList:		//=============================   
					switch(keyPress) {
						case kpOk:			//------------ key OK
							{
								uint16_t freq = eeprom_read_word(&eeSettings.radio.presets[subMode]);
								if (freq) {
									rda5807mSetFrequency(freq);
									currentChannel = subMode;
									newMode = mShowRadio;
								} else currentChannel = 0xFF;
								displaySetChannel(currentChannel);
							}
							break;
						case kpOkLong:		//------------ key OK (long press)
							radioPowerControl = setRadioPowerOff;
							newMode = mRadioTurnOff;
							break;
						case kpLeftLong:	//------------ key LEFT (long press)
							isPresetSubMenu = 0;
							newMode = mShowRadio;
							currentChannel = findNearChannel(rda5807mGetFrequency(), 1);
							displaySetChannel(currentChannel);
							break;
						case kpLeft:		//------------ key LEFT
							menuChange(&mmPresets);
							newMode = mMainMenu;
							break;
						case kpRight:		//------------ key RIGHT
						case kpRightLong:	//------------ key RIGHT (long press)
							menuChange(&mmChSaveDel);
							isPresetSubMenu = 1;
							newMode = mMainMenu;
							break;
						case kpUp:			//------------ key UP
							if (subMode) subMode--;
							else subMode = FM_PRESETS-1;
							displayMenuIsChange();
							break;
						case kpDown:		//------------ key DOWN
							if (subMode<FM_PRESETS-1) subMode++;
							else subMode = 0;
							displayMenuIsChange();
							break;
					} // switch (keyPress)
					break;
				case mMainMenu:			//=============================  
					switch(keyPress) {
						case kpOkLong:		//------------ key OK (long press)
							radioPowerControl = setRadioPowerOff;
							newMode = mRadioTurnOff;
							break;
						case kpOk:			//------------ key OK
							i = MENU_COMMAND;
							switch (/*(menuCommandType)*/i) {
								case mcmdGoPresets:
									goto goPresetsMode;
									//break;
								case mcmdMonoStereo:
									i = !eeprom_read_byte(&eeSettings.radio.stereo);
									eeprom_update_byte(&eeSettings.radio.stereo, i);
									EEPROM_UpdateCRC();
									rda5807mSetAudioMode( i ? audioStereo : audioForceMono);
									displayMenuIsChange();
									break;
								case mcmdLoudness:
									i = !eeprom_read_byte(&eeSettings.radio.bass);
									eeprom_update_byte(&eeSettings.radio.bass, i);
									EEPROM_UpdateCRC();
									rda5807mSetBassBoost( i ? bassboostOn : bassBoostOff);
									displayMenuIsChange();
									break;
								case mcmdRDSmode:
									i = !eeprom_read_byte(&eeSettings.radio.showRds);
									eeprom_update_byte(&eeSettings.radio.showRds, i);
									EEPROM_UpdateCRC();
									displayMenuIsChange();
									break;
								case mcmdKeySound1:
								case mcmdKeySound0:
									eeprom_update_byte(&eeSettings.system.keySound, i-mcmdKeySound0);
									EEPROM_UpdateCRC();
									menuChange(MENU_PARENT); 
									displayMenuIsChange();
									break;
								case mcmdKeyActOff:
								case mcmdKeyActOn:
									eeprom_update_byte(&eeSettings.system.keyAction, i-mcmdKeyActOff);
									EEPROM_UpdateCRC();
									menuChange(MENU_PARENT);
									displayMenuIsChange();
									break;
								case mcmdSetBkLight:
									eeprom_update_byte(&eeSettings.system.lightTime, onScreenEditVal);
									EEPROM_UpdateCRC();
									menuChange(MENU_PARENT);
									displayMenuIsChange();
									break;
								case mcmdSetBatVolt:
									eeprom_update_byte(&eeSettings.battery.bandGap, batteryGetReferenceVoltage(onScreenEditVal));
									EEPROM_UpdateCRC();
									menuChange(MENU_PARENT);
									displayMenuIsChange();
									updateBatteryFilter = 1;
									break;
								case mcmdSetLowBat:
									eeprom_update_word(&eeSettings.battery.lowVoltage, onScreenEditVal);
									EEPROM_UpdateCRC();
									menuChange(MENU_PARENT);
									displayMenuIsChange();
									break;
								case mcmdDevClear:
									{
										void* ee = &eeSettings;
										for (i=0; i<sizeof(settingsType); i++) eeprom_update_byte(ee++,0xFF);
									}
									// break; -    
								case mcmdDevRestart:
									cli();
									WDTCSR = 1<<WDE | 1<<WDP2;//     0.25 
									while(1);
									break;

								case mcmdMoveChUp:
									if (subMode) {
										uint16_t prev = eeprom_read_word(&eeSettings.radio.presets[subMode-1]);
										uint16_t curr = eeprom_read_word(&eeSettings.radio.presets[subMode  ]);
										eeprom_update_word(&eeSettings.radio.presets[subMode-1], curr);
										eeprom_update_word(&eeSettings.radio.presets[subMode  ], prev);
										EEPROM_UpdateCRC();
										subMode--;
										newMode = mPresetList;
										isPresetSubMenu = 0;
									}
									break;
								case mcmdMoveChDn:
									if (subMode<FM_PRESETS-1) {
										uint16_t next = eeprom_read_word(&eeSettings.radio.presets[subMode+1]);
										uint16_t curr = eeprom_read_word(&eeSettings.radio.presets[subMode  ]);
										eeprom_update_word(&eeSettings.radio.presets[subMode+1], curr);
										eeprom_update_word(&eeSettings.radio.presets[subMode  ], next);
										EEPROM_UpdateCRC();
										subMode++;
										newMode = mPresetList;
										isPresetSubMenu = 0;
									}
									break;
								case mcmdChInsLine:
									if (subMode<FM_PRESETS-1) {
										uint8_t i = FM_PRESETS-1;
										while (i>subMode) {
											uint16_t rec = eeprom_read_word(&eeSettings.radio.presets[i-1]);
											eeprom_update_word(&eeSettings.radio.presets[i], rec);
											i--;
										}
									}
									eeprom_update_word(&eeSettings.radio.presets[subMode], 0);
									EEPROM_UpdateCRC();
									newMode = mPresetList;
									isPresetSubMenu = 0;
									break;
								case mcmdChDelLine:
									if (subMode<FM_PRESETS-1) {
										uint8_t i = subMode;
										while (i<FM_PRESETS-1) {
											uint16_t rec = eeprom_read_word(&eeSettings.radio.presets[i+1]);
											eeprom_update_word(&eeSettings.radio.presets[i], rec);
											i++;
										}
									}
									eeprom_update_word(&eeSettings.radio.presets[FM_PRESETS-1], 0);
									EEPROM_UpdateCRC();
									newMode = mPresetList;
									isPresetSubMenu = 0;
									break;
								case mcmdChSaveDel:
									{
										uint16_t freq = eeprom_read_word(&eeSettings.radio.presets[subMode]);
										if (freq) freq = 0;
										else freq = rda5807mGetFrequency();
										eeprom_update_word(&eeSettings.radio.presets[subMode], freq);
										EEPROM_UpdateCRC();
									}
									newMode = mPresetList;
									isPresetSubMenu = 0;
									break;
								case mcmdChSetFreq:
									eeprom_update_word(&eeSettings.radio.presets[subMode], onScreenEditVal*5);
									EEPROM_UpdateCRC();
									newMode = mPresetList;
									isPresetSubMenu = 0;
									break;
								case mcmdAutoScan:	 
									newMode = mAutoScan;
									//   
									for (i=0; i<FM_PRESETS; i++) eeprom_update_word(&eeSettings.radio.presets[i],0);
									//   
									rda5807mSetFrequency(8700);
									rda5807mSearch(tuneUp);
									scanChannel = 0;
									stopScan = 0;
									displaySetScanProgress(0);
									break;
								case mcmdSetSleep:
									if (sleepMin < 5) sleepMin = 5;
									else if (sleepMin < 10) sleepMin = 10;
									else if (sleepMin < 15) sleepMin = 15;
									else if (sleepMin < 20) sleepMin = 20;
									else if (sleepMin < 30) sleepMin = 30;
									else if (sleepMin < 60) sleepMin = 60;
									else if (sleepMin < 90) sleepMin = 90;
									else sleepMin = 0;
									sleepSec = 60;
									displaySetSleepTime(sleepMin);
									break;
							} // switch i (menu command)
							isScreenEdit = 0;	 
							break;
						case kpLeftLong:	//------------ key LEFT (long press)		 
							isPresetSubMenu = 0;
							newMode = mShowRadio;
							currentChannel = findNearChannel(rda5807mGetFrequency(), 1);
							displaySetChannel(currentChannel);
							break;
						case kpLeft:		//------------ key LEFT
							isScreenEdit = 0;
							i = MENU_COMMAND;
							if (i == mcmdChSetFreq) {
								//       
								//      
								rda5807mSetFrequency(saveFreq);
							}
							if (menuChange(MENU_PARENT)) {
								displayMenuIsChange();
								break;
							}
							if (isPresetSubMenu) {//   ,    ,   ,   
								newMode = mPresetList;
								isPresetSubMenu = 0;
							} else {
								newMode = mShowRadio;	
								currentChannel = findNearChannel(rda5807mGetFrequency(), 1);
								displaySetChannel(currentChannel);
							}
							break;
						case kpRight:		//------------ key RIGHT
						case kpRightLong:	//------------ key RIGHT (long press)
							isScreenEdit = 0;
							if (menuChange(MENU_CHILD)) displayMenuIsChange();
							i = MENU_COMMAND;
							//   ,           
							switch (/*(menuCommandType)*/i) {
								case mcmdGoPresets:
		goPresetsMode:				//    -   				
									newMode = mPresetList;
									//      -  .    .
									if (currentChannel != 0xFF) subMode = currentChannel;
									else {
										subMode = 0;
										while (subMode < FM_PRESETS-2  &&  eeprom_read_word(&eeSettings.radio.presets[subMode])) subMode++;
									}
									break;
								case mcmdKeySound1:
									//        
									if ( !eeprom_read_byte(&eeSettings.system.keySound) ) menuChange(MENU_NEXT);
									break;
								case mcmdKeyActOff:
									if ( eeprom_read_byte(&eeSettings.system.keyAction) ) menuChange(MENU_NEXT);
									break;
								case mcmdSetBkLight:
									isScreenEdit = 1;
									rollingValue = 0;
									onScreenEditVal = eeprom_read_byte(&eeSettings.system.lightTime);
									onScreenEditMin = 2;
									onScreenEditMax = 30;
									displaySetScreenEditValue(onScreenEditVal);
									break;
								case mcmdSetBatVolt:
									isScreenEdit = 1;
									rollingValue = 0;
									onScreenEditVal = batVoltage;
									onScreenEditMin = 300;
									onScreenEditMax = 490;
									displaySetScreenEditValue(onScreenEditVal);
									break;
								case mcmdSetLowBat:
									isScreenEdit = 1;
									rollingValue = 0;
									onScreenEditVal = eeprom_read_word(&eeSettings.battery.lowVoltage);
									onScreenEditMin = 290;
									onScreenEditMax = 360;
									displaySetScreenEditValue(onScreenEditVal);
									break;
								case mcmdSetTime:
									newMode = mSetTime;
									subMode = 0;
									timeEditor.hour = time.hour;
									timeEditor.min = time.min;
									break;
								case mcmdChSetFreq:
									saveFreq = rda5807mGetFrequency();
									isScreenEdit = 1;
									rollingValue = 1;
									onScreenEditMin = 8700/5;
									onScreenEditMax = 10800/5;
									{
										uint16_t freq = eeprom_read_word(&eeSettings.radio.presets[subMode]);
										if (freq) onScreenEditVal = freq/5;
										else onScreenEditVal = rda5807mGetFrequency()/5;
									}
									displaySetScreenEditValue(onScreenEditVal);
									break;
							} // switch i (menu command)
							break;
						case kpUp:			//------------ key UP
							if (isScreenEdit) {
								if (onScreenEditVal<onScreenEditMax) {
									onScreenEditVal++;
									displaySetScreenEditValue(onScreenEditVal);
								} else if (rollingValue) {
									onScreenEditVal = onScreenEditMin;
									displaySetScreenEditValue(onScreenEditVal);
								}
								if (MENU_COMMAND == mcmdChSetFreq) rda5807mSetFrequency(onScreenEditVal*5);
								break;
							}
							if (menuChange(MENU_PREVIOUS)) displayMenuIsChange();
							break;
						case kpDown:		//------------ key DOWN
							if (isScreenEdit) {
								if (onScreenEditVal>onScreenEditMin) {
									onScreenEditVal--;
									displaySetScreenEditValue(onScreenEditVal);
								} else if (rollingValue) {
									onScreenEditVal = onScreenEditMax;
									displaySetScreenEditValue(onScreenEditVal);
								}
								if (MENU_COMMAND == mcmdChSetFreq) rda5807mSetFrequency(onScreenEditVal*5);
								break;
							}
							if (menuChange(MENU_NEXT)) displayMenuIsChange();
							break;
					} // switch (keyPress) 
					break;
			} // switch (mode)
		} // if keyPress != kpNoPress
		
		//-------------------------------------------   
		if (modeTimeOut && mode == newMode) {
			if (clk>modeTimeOut) modeTimeOut = 0;
			else modeTimeOut -= clk;
			if (!modeTimeOut) { //  
				switch (mode) {
					case mSplashScreen:		//=============================     
						newMode = mRadioTurnOff;
						break;
					case mShowTime:			//=============================   ( )
						newMode = mTimeGoSleep;
						break;
					case mTimeGoSleep:		//=============================      
					case mRadioTurnOff:		//=============================  
						newMode = mSleep;
						break;				
					case mRadioTurnOn:		//=============================    
						//   ,      ,     
						radioPowerControl = setRadioPowerOn;
						newMode = mShowRadio;
						break;
					case mRadioScanUp:		//=============================   /
					case mRadioScanDown:	//=============================   /
					case mMainMenu:			//=============================  
					case mPresetList:		//=============================   
						currentChannel = findNearChannel(rda5807mGetFrequency(), 1);
						displaySetChannel(currentChannel);
						newMode = mShowRadio;
						break;				
					case mAutoScan:			//=============================  
						newMode = mPresetList;
						break;
					case mShowRadio:		//=============================   
						newMode = mRadioGoSleep;
						break;
					case mRadioGoSleep:		//=============================      
						newMode = mRadioSleep;
						break;
					case mNoHeadphones:		//=============================  
					case mBatteryLow:		//=============================   
						radioPowerControl = setRadioPowerOff;
						newMode = mRadioTurnOff;
						break;
					case mErrNoFm:			//=============================  -   
						newMode = mErrNoFmSleep;
						break;
					case mRadioSleep:		//=============================    
					case mErrNoFmSleep:		//=============================    " "
					case mSleep:			//=============================   ( )
						break;
					case mSetTime:			//=============================  
						newMode = mMainMenu;
						menuChange(&mmTimeSetup);
						break;
				} // switch (mode)
				
			} // if (!modeTimeOut)
		} // if (modeTimeOut && mode == newMode)
		
		//-------------------------------------------  
		if (reloadTimeOut || mode != newMode) {
			switch (newMode) {
				case mSleep:			//=============================   ( )
				case mRadioSleep:		//=============================    
				case mErrNoFmSleep:		//=============================    " "
					modeTimeOut = 0;
					break;
				case mShowRadio:		//=============================   
					modeTimeOut = (uint16_t)eeprom_read_byte(&eeSettings.system.lightTime) * 1000;
					break;
				case mSplashScreen:		//=============================     
				case mRadioTurnOn:		//=============================    
				case mRadioScanUp:		//=============================   /
				case mRadioScanDown:	//=============================   /
				case mRadioTurnOff:		//=============================  
				case mAutoScan:			//=============================  
				case mBatteryLow:		//=============================   
				case mNoHeadphones:		//=============================  
				case mRadioGoSleep:		//=============================      
				case mTimeGoSleep:		//=============================      
					modeTimeOut = 60000; //  ,        
					break;
				case mSetTime:			//  
					modeTimeOut = 65535u; // 65 
					break;
				case mMainMenu:			//=============================  
				case mPresetList:		//=============================   
					modeTimeOut = 30000;
					break;
				case mErrNoFm:			//=============================  -   
				case mShowTime:			//=============================   ( )
					modeTimeOut = TEMPORARY_MODES_TIME;		//    
					break;
			} // switch (mode)
		} // if (reloadTimeOut || mode != newMode)
		
		
		//-------------------------------------------   
		if (radioPowerControl == setRadioPowerOff ) {
			//  
			eeprom_update_word(&eeSettings.radio.lastFreq, rda5807mGetFrequency());
			eeprom_update_byte(&eeSettings.radio.lastVol, radioVolume);
			EEPROM_UpdateCRC();
			extAmpControlSetState(0);
			rda5807mSetPowerMode(powerOff);
			phonesDisableCheckPresent();
			sleepMin = 0;
			sleepSec = 0;
			radioPower = 0;
			displaySetSleepTime(0);
		}

		if (radioPowerControl == setRadioPowerOn ) {
			//  
			rda5807mSetBand(bandEuropeUsa);
			rda5807mSetFreqStep(frequencyStep50kHz);
			rda5807mSetPowerMode(powerOn);
			phonesEnableCheckPresent();
			extAmpControlSetState(1);
			rda5807mSetAudioMode(eeprom_read_byte(&eeSettings.radio.stereo) ? audioStereo : audioForceMono);
			rda5807mSetBassBoost(eeprom_read_byte(&eeSettings.radio.bass) ? bassboostOn : bassBoostOff);
			radioVolume = eeprom_read_byte(&eeSettings.radio.lastVol);
			if (radioVolume) {
				rda5807mSetMuteMode(muteOff);
				rda5807mSetVolume(radioVolume-1);
			} else {
				rda5807mSetMuteMode(muteOn);
				rda5807mSetVolume(0);
			}
			uint16_t freq = eeprom_read_word(&eeSettings.radio.lastFreq);
			rda5807mSetFrequency(freq);
			displaySetFrequency(freq);
			displaySetVolumeVal(radioVolume);
			currentChannel = findNearChannel(freq, 1);
			displaySetChannel(currentChannel);
			sleepMin = 0;
			sleepSec = 0;
			radioPower = 1;
			displaySetSleepTime(0);
		}

		//-------------------------------------------  
		//========   
		if (mode == mShowRadio ||
			mode == mRadioScanUp ||
			mode == mRadioScanDown ||
			mode == mAutoScan ) {
			static uint8_t pollCounter = 200;
			if ( mode != newMode ) pollCounter = 190;  // 10        
			if (pollCounter) pollCounter--;
			else {
				pollCounter = 190; // 10 ms    
				//     
				rda5807mReadState();
				displaySetFrequency(rda5807mState.frequency);
				displaySetReceiveFlags(rda5807mState.strength, rda5807mState.stereo);
				if ( mode == mShowRadio) {
					displaySetRdsStation(rda5807mState.rdsReady && eeprom_read_byte(&eeSettings.radio.showRds), rda5807mState.rdsStation);
				}
				if ( mode == mAutoScan) {
					if (!stopScan && rda5807mState.seekComplete) {
						eeprom_update_word(&eeSettings.radio.presets[scanChannel], rda5807mState.frequency);
						scanChannel++;
						if (scanChannel<FM_PRESETS) rda5807mSearch(tuneUp);
						displaySetScanProgress(scanChannel);
					}
					if (stopScan || rda5807mState.seekFail || scanChannel == FM_PRESETS) {
						EEPROM_UpdateCRC();
						subMode = 0;
						modeTimeOut = 1;
						if (scanChannel) {
							currentChannel = 0;
							rda5807mSetFrequency(eeprom_read_word(&eeSettings.radio.presets[0]));
						} else {
							currentChannel = 0xFF;
							rda5807mSetFrequency(8700);
						}
					} 
				}
				if ( mode == mRadioScanUp || mode == mRadioScanDown ) {
					if (rda5807mState.seekComplete || rda5807mState.seekFail) {
						modeTimeOut = 1;
						currentChannel = findNearChannel(rda5807mState.frequency, 1);
						displaySetChannel(currentChannel);
					}
				}
			}	// if pollCounter .. else
		} // if ( mode == || ...	  
		
		
		//-------------------------------------------  
		mode = newMode;
		
		//------------------------------------------- 
		drawDisplay(mode);   
		if (drawProcessIsEnd()) modeTimeOut = 1;
		
		//-------------------------------------------  
		if (keyWatchCounter) keyWatchCounter--;
		if (mode == mRadioSleep || mode == mSleep || mode == mErrNoFmSleep) {
			if (!keyPress && !keyWatchCounter && !soundPlayCounter) { //      ...
				if (mode != mRadioSleep) {
					keysEnableWakeUp(KEY_CENTER);
					phonesDisableCheckPresent();
				} else keysEnableWakeUp(KEYS_MASK);
				cli();
				keyWake = 0;
				time.changed = 0;
				sei();
				// !!!!
				goSleep();
				//   .  ,  ....
				//    -  ,  
				//      -     .
				if (keyWake) resetKeysBySleep = 1;
			}	// if (    )
		}// if sleep modes
    } // super LOOP
	
}  // The END

