/*
 * GraphLib.c
 *
 * Created: 04.10.2018 10:44:37
 *  Author: GoldenAndy
 */ 

#include "GraphLib.h"
#include <avr/pgmspace.h>

static unsigned int drawColor  = 0;
static unsigned int bgColor   = 0;

static coordType lastX = 0;
static coordType lastY = 0;


void setDrawColor(unsigned int color){
	drawColor = color;
}

void setBackgroundColor(unsigned int color){
	bgColor = color;
}

unsigned int convertRgbTo565color(unsigned char r, unsigned char g, unsigned char b){
	//unsigned int color = b >> 3;
	//color |= ((unsigned int)g & 0xFC) << 3;
	//color |= ((unsigned int)r & 0xF8) << 8;
	//return color;
	return MAKE565_FROM_RGB(r,g,b);
}



void normalizeCoord(coordType* c1, coordType* c2){
	if (*c1 > *c2) {
		coordType c = *c1;
		*c1 = *c2;
		*c2 = c;
	}
}


void line (coordType x1, coordType y1, coordType x2, coordType y2){
	lastX = x2;
	lastY = y2;
	//  
	if (x1==x2) {
		normalizeCoord(&y1, &y2);
		#ifdef DISPLAY_USE_AREA
			displayFillArea(x1,y1,x2,y2,drawColor);
		#else
			while(1) {
				displayPutPixel(x1,y1,drawColor);
				if (y1==y2) break;
				y1++;
			}
		#endif
		return;
	}
	//  
	if (y1==y2) {
		normalizeCoord(&x1, &x2);
		#ifdef DISPLAY_USE_AREA
		displayFillArea(x1,y1,x2,y2,drawColor);
		#else
		while(1) {
			displayPutPixel(x1,y1,drawColor);
			if (x1==x2) break;
			x1++;
		}
		#endif
		return;
	}
	// -  - 
	
	const signed int deltaX=(x2 > x1) ? (x2-x1) : (x1-x2);
	const signed int deltaY=(y2 > y1) ? (y2-y1) : (y1-y2);
	const signed char signX = x1 < x2 ? 1 : -1;
	const signed char signY = y1 < y2 ? 1 : -1;
	//
	int error = deltaX - deltaY;
	//
	displayPutPixel(x2, y2, drawColor);
	while(x1 != x2 || y1 != y2)
	{
		displayPutPixel(x1, y1, drawColor);
		const int error2 = error * 2;
		//
		if(error2 > -deltaY)
		{
			error -= deltaY;
			x1 += signX;
		}
		if(error2 < deltaX)
		{
			error += deltaX;
			y1 += signY;
		}
	}
}

void lineTo (coordType x2, coordType y2){
	line(lastX,lastY,x2,y2);	
}

void rectangle (coordType x1, coordType y1, coordType x2, coordType y2, fillModeType fill){
	lastX = x2;
	lastY = y2;
	if(fill==fmNoFill) {
		line(x1, y1, x2, y1);
		line(x2, y1, x2, y2);
		line(x1, y1, x1, y2);
		line(x1, y2, x2, y2);
		return;
	}
	//Filled
	#ifdef DISPLAY_USE_AREA
		normalizeCoord(&x1, &x2);
		normalizeCoord(&y1, &y2);
		displayFillArea(x1,y1,x2,y2,drawColor);
	#else 
		normalizeCoord(&y1, &y2);
		coordType i=y1;
		while(1) {
			line(x1, i, x2, i);
			if (i==y2) break;
			i++;
		}
	#endif
}


void circle (coordType x0, coordType y0, coordType R, fillModeType fill){
	lastX = x0;
	lastY = y0;
	unsigned char x=0;
	short int y=R;
	short int delta=(1-2*R);
	short int error=0;
	while(y>=0)
	{
		if(fill==fmNoFill)
		{
			displayPutPixel(x0+x-1, y0+y, drawColor);
			displayPutPixel(x0+x-1, y0-y, drawColor);
			displayPutPixel(x0-x+2, y0+y, drawColor);
			displayPutPixel(x0-x+2, y0-y, drawColor);
		}
		else
		{
			#ifdef DISPLAY_USE_AREA
				displayFillArea(x0+x-1,y0-y,x0+x-1,y0+y,drawColor);
				displayFillArea(x0-x+2,y0-y,x0-x+2,y0+y,drawColor);
			#else
				line(x0+x-1,y0-y,x0+x-1,y0+y);
				line(x0-x+2,y0-y,x0-x+2,y0+y);
			#endif
		}
		error = 2*(delta+y)-1;														
		if((delta<0)&&(error<=0))
		{
			x++;
			delta+=(2*x + 1);
			continue;
		}
		error=2*(delta-x)-1;
		if((delta > 0)&&(error>0))
		{
			y--;
			delta+=(1-2*y);
			continue;
		}
		x++;
		delta+=(2*(x - y));
		y--;
	}
}

#ifdef DISPLAY_USE_FONTS

	typedef enum {
	  fcstDigits			= 0,	//    
	  fcstDigitsSymbols		= 1,	// , 
	  fcstEngSet			= 2,	// ,   (.  .)
	  fcstEngRusSet			= 3,	// , .  .  (.  .)
	  fcstEngUaSet			= 4,	// , .  +  (.  .)
	  fcstFullMapSet		= 5		//   
			} fontCharSetMapType;
	
	#ifndef NO_FONT_EFFECTS
		static struct {
					displayFontType			fontType;
					unsigned char			symWidth;
					unsigned char			symHeight;
					fontCharSetMapType		FontSet;
					unsigned char			bitDir;
					unsigned char			dataSize;
					void*					font;
					unsigned char			symbolsInterval;
					textOutModeType			textOutMode;
					textOrientationType		textOrientation;
					textMirrorType			textMirror;
					fontColorMapType		fontColorMap;
					textIntervalFillType	textIntervalFill; 
				} currentFontSettings = {	dummyFontFt,		//	displayFontType			fontType;
											0, 					//	unsigned char			symWidth;
											0, 					//	unsigned char			symHeight;
											fcstDigits, 		//	fontCharSetMapType		FontSet;
											0, 					//	unsigned char			bitDir;
											1, 					//	unsigned char			dataSize;
											0, 					//	void*					font;
											1, 					//	unsigned char			symbolsInterval;
											tomNormal, 			//	textOutModeType			textOutMode;
											toNormal, 			//	textOrientationType		textOrientation;
											tmNormal, 			//	textMirrorType			textMirror;
											fcmUseDrawColor, 	//	fontColorMapType		fontColorMap;
											tifNoFill 			//	textIntervalFillType	textIntervalFill;
											};
	#else // ifndef NO_FONT_EFFECTS
		// defined NO_FONT_EFFECTS									
		static struct {
					displayFontType			fontType;
					unsigned char			symWidth;
					unsigned char			symHeight;
					fontCharSetMapType		FontSet;
					unsigned char			bitDir;
					unsigned char			dataSize;
					void*					font;
					unsigned char			symbolsInterval;
					//textOutModeType			textOutMode;
					//textOrientationType		textOrientation;
					//textMirrorType			textMirror;
					//fontColorMapType		fontColorMap;
					textIntervalFillType	textIntervalFill; 
				} currentFontSettings = {	dummyFontFt,		//	displayFontType			fontType;
											0, 					//	unsigned char			symWidth;
											0, 					//	unsigned char			symHeight;
											fcstDigits, 		//	fontCharSetMapType		FontSet;
											0, 					//	unsigned char			bitDir;
											1, 					//	unsigned char			dataSize;
											0, 					//	void*					font;
											1, 					//	unsigned char			symbolsInterval;
											//tomNormal, 			//	textOutModeType			textOutMode;
											//toNormal, 			//	textOrientationType		textOrientation;
											//tmNormal, 			//	textMirrorType			textMirror;
											//fcmUseDrawColor, 	//	fontColorMapType		fontColorMap;
											tifNoFill 			//	textIntervalFillType	textIntervalFill;
											};
	#endif // ifndef NO_FONT_EFFECTS
	
	#ifndef NO_FONT_EFFECTS
		static unsigned int* fontColorMapArr = 0;
	#endif

	//  
	#ifdef USE_FONT_NOKIA_5X7
		#include "FONTS/5x7_Nokia.h"
		const unsigned char nokia5x7font[]  PROGMEM = FONT_5X7_NOKIA;
	#endif

	#ifdef USE_FONT_NOKIA_DIGITS_5X7
		#include "FONTS/5x7_Nokia_Digits.h"
		const unsigned char nokiaDigits5x7font[]  PROGMEM = FONT_5X7_NOKIA_DIGITS;
	#endif

	#ifdef USE_FONT_SINCLAIR_7X7
		#include "FONTS/7x7_Sinclair.h"
		const unsigned char sinclair7x7font[]  PROGMEM = FONT_7X7_SINCLAIR;
	#endif

	#ifdef USE_FONT_TIMES_7X11
		#include "FONTS/7x11_Times.h"
		const unsigned char times7x11font[]  PROGMEM = FONT_7X11_TIMES;
	#endif

	#ifdef USE_FONT_BIGDIGITS_15X23 
		#include "FONTS/15x23_digits.h"
		const unsigned char digits15x23font[]  PROGMEM = FONT_15X23_BIGDIGITS;
	#endif 


	#ifdef USE_FONT_ARIAL_14X21 
		#include "FONTS/14x21_Arial.h"
		const unsigned char arial14x21font[]  PROGMEM = FONT_14X21_ARIAL_ROUND;
	#endif 

	#ifdef USE_FONT_HALLVETICA_9X15 
		#include "FONTS/9x15_Hallfetica.h"
		const unsigned char hallfetica9x15font[]  PROGMEM = FONT_9X15_HALLFETICA_NORMAL;
	#endif 
	
	#ifdef USE_FONT_COND_HALLVET_7X15 
		#include "FONTS/7x15_Hallfetica_condensed.h"
		const unsigned char condHallfetica7x15font[]  PROGMEM = FONT_7X15_HALLFETICA_CONDENSE;
	#endif 

	#ifdef USE_FONT_DIGITS_7X11 
		#include "FONTS/7x11_DigitsSymbols.h"
		const unsigned char digits7x11font[]  PROGMEM = FONT_7X11_DIGITSSYMBOLS;
	#endif 

	#ifdef USE_FONT_HALLV_T_DIGS_9X15 
		#include "FONTS/9x15_HallfeticaDigitsTemperature.h"
		const unsigned char hallfe_digs9x15font[]  PROGMEM = FONT_9X15_HALLFETICA_TEMPERAT;
	#endif 

	#ifdef USE_FONT_CRYSTAL_9X15 
		#include "FONTS/9x15_crystal.h"
		const unsigned char crystal_9x15font[]  PROGMEM = FONT_9X15_CRYSTAL;
	#endif 
	#ifdef USE_FONT_NIXIE_9X15 
		#include "FONTS/9x15_Nixie.h"
		const unsigned char nixie_9x15font[]  PROGMEM = FONT_9X15_NIXIE;
	#endif 

	#ifdef USE_FONT_NIXIE_11X19 
		#include "FONTS/11x19_Nixie.h"
		const unsigned char nixie_11x19font[]  PROGMEM = FONT_11X19_NIXIE;
	#endif 

	#ifdef USE_FONT_NIXIE_DIGITS_11X19 
		#include "FONTS/11x19_NixieDigits.h"
		const unsigned char nixie_digs11x19font[]  PROGMEM = FONT_11X19_NIXIEDIGITS;
	#endif 

	#ifdef USE_FONT_BIG_18X28 
		#include "FONTS/18x28_big.h"
		const unsigned char big18x28font[]  PROGMEM = FONT_18X28_BIG;
	#endif

	#ifdef USE_FONT_BIG_20X32_DIGITS 
		#include "FONTS/20x32_big_digits.h"
		const unsigned char big20x32fontDigits[]  PROGMEM = FONT_20X32_BIGDIGITS;
	#endif

	#ifdef USE_FONT_BIG_METEO_14X28 
		#include "FONTS/14x28_bigTemperature.h"
		const unsigned char bigMeteo14x28font[]  PROGMEM = FONT_14X28_BIG_TEMPERATURE;
	#endif

	#ifdef USE_FONT_BIG_CRYSTAL_18X28 
		#include "FONTS/18x28_big_crystal.h"
		const unsigned char bigCrystal18x28font[]  PROGMEM = FONT_18X28_BIG_CRYSTAL;
	#endif	

	#ifdef USE_FONT_BCRYST_METEO_14X28 
		#include "FONTS/14x28_bigCrystalTemperature.h"
		const unsigned char bigCrystalMeteo14x28font[]  PROGMEM = FONT_14X28_BIG_CRYSTAL_TEMPER;
	#endif	
	
	const unsigned int fontData [] PROGMEM = {	
		
		#ifdef USE_FONT_NOKIA_5X7
		(unsigned int)&nokia5x7font,
		#endif
	
		#ifdef USE_FONT_NOKIA_DIGITS_5X7
		(unsigned int)&nokiaDigits5x7font,
		#endif
		
		#ifdef USE_FONT_SINCLAIR_7X7
		(unsigned int)&sinclair7x7font,
		#endif
	
		#ifdef USE_FONT_TIMES_7X11
		(unsigned int)&times7x11font,
		#endif
	
		#ifdef USE_FONT_BIGDIGITS_15X23
		(unsigned int)&digits15x23font,
		#endif
		
		#ifdef USE_FONT_ARIAL_14X21
		(unsigned int)&arial14x21font,
		#endif
		
		#ifdef USE_FONT_HALLVETICA_9X15
		(unsigned int)&hallfetica9x15font,
		#endif
		
		#ifdef USE_FONT_COND_HALLVET_7X15
		(unsigned int)&condHallfetica7x15font,
		#endif
		
		#ifdef USE_FONT_DIGITS_7X11
		(unsigned int)&digits7x11font,
		#endif	
		
		#ifdef USE_FONT_HALLV_T_DIGS_9X15
		(unsigned int)&hallfe_digs9x15font,
		#endif
		
		#ifdef USE_FONT_CRYSTAL_9X15
		(unsigned int)&crystal_9x15font,
		#endif

		#ifdef USE_FONT_NIXIE_9X15
		(unsigned int)&nixie_9x15font,
		#endif

		#ifdef USE_FONT_NIXIE_11X19
		(unsigned int)&nixie_11x19font,
		#endif

		#ifdef USE_FONT_NIXIE_DIGITS_11X19
		(unsigned int)&nixie_digs11x19font,
		#endif

		#ifdef USE_FONT_BIG_18X28
		(unsigned int)&big18x28font,
		#endif
		
		#ifdef USE_FONT_BIG_20X32_DIGITS
		(unsigned int)&big20x32fontDigits,
		#endif
		
		#ifdef USE_FONT_BIG_METEO_14X28
		(unsigned int)&bigMeteo14x28font,
		#endif
		
		#ifdef USE_FONT_BIG_CRYSTAL_18X28
		(unsigned int)&bigCrystal18x28font,
		#endif		
		
		#ifdef USE_FONT_BCRYST_METEO_14X28
		(unsigned int)&bigCrystalMeteo14x28font,
		#endif		
		
		};
	
		#ifdef FONT_LEFT_MSB
		#define bdLeftMsb		0
		#endif
	
		#ifdef FONT_BOTTOM_MSB
		#define bdBottomMsb		1
		#endif
	
		#ifdef FONT_TOP_MSB
		#define bdTopMsb		2
		#endif
	
		#ifdef FONT_BIT_FLOW
		#define bdBitFlow		3
		#endif
		
	#ifndef NO_FONT_EFFECTS	
		void setTextOutMode(textOutModeType textOutMode){
			currentFontSettings.textOutMode = textOutMode;
		}
		
		void setTextOrientation(textOrientationType textOrientation){
			currentFontSettings.textOrientation = textOrientation;
		}
		
		void setTextMirrorMode(textMirrorType textMirror){
			currentFontSettings.textMirror = textMirror;
		}
			
		void setTextColorMap(fontColorMapType fontColorMapMode, unsigned int fontColorMap[]){
			currentFontSettings.fontColorMap = fontColorMapMode;
			fontColorMapArr = fontColorMap;	
		}
	#endif // ifndef NO_FONT_EFFECTS

	void setTextSymbolsInterval(unsigned char symbolsInterval){
		currentFontSettings.symbolsInterval = symbolsInterval;
	}
	
	void setTextIntervalFillType(textIntervalFillType textIntervalFillMode){
		currentFontSettings.textIntervalFill = textIntervalFillMode;
	}
		
	static void specialPutPixel(int dx,int dy,coordType x0,coordType y0, unsigned int color){
		#ifndef NO_FONT_EFFECTS
			if (currentFontSettings.textMirror == tmHMirror)  dx = -dx;
			if (currentFontSettings.textMirror == tmVMirror)  dy = -dy;
			switch (currentFontSettings.textOrientation) {
				case toNormal :	   
					displayPutPixel(x0+dx,y0+dy,color);
					break;
				case to90 :
					displayPutPixel(x0+dy,y0-dx,color);
					break;
				case to180 :
					displayPutPixel(x0-dx,y0-dy,color);
					break;
				case to270 :
					displayPutPixel(x0-dy,y0+dx,color);
					break;	  
			}	//switch	  
		#else // ifndef NO_FONT_EFFECTS
			// defined NO_FONT_EFFECTS
			#ifdef DISPLAY_USE_AREA
				displaySendColor(color);
			#else
				displayPutPixel(x0+dx,y0+dy,color);
			#endif // ifdef DISPLAY_USE_AREA
		#endif // ifndef NO_FONT_EFFECTS
	}
				
	void putSymbol (coordType x, coordType y, char symbol, displayFontType fontType){
		if (fontType != currentFontSettings.fontType) {
			currentFontSettings.fontType = fontType;
			unsigned char* fontPtr = (unsigned char*)pgm_read_word(&fontData[fontType]);
			currentFontSettings.symWidth  = pgm_read_byte(fontPtr++);
			currentFontSettings.symHeight = pgm_read_byte(fontPtr++);
			currentFontSettings.FontSet   = pgm_read_byte(fontPtr++);
			currentFontSettings.bitDir	  = pgm_read_byte(fontPtr++);
			currentFontSettings.dataSize  = pgm_read_byte(fontPtr++);
			currentFontSettings.font = fontPtr;
		}
		//       ,       
		unsigned int myIndex = 0;
		switch (currentFontSettings.FontSet) {
			#ifdef FONT_DIGITS_MAP
			case fcstDigits: 
				//  -    
				if (symbol==' ') myIndex = 0;
				else if (symbol >= '0' && symbol <='9') myIndex = symbol-'0'+1;
				break;
				#endif
				
			#ifdef FONT_DIGITS_SYMBOLS_MAP
			case fcstDigitsSymbols: 
				//  -   
				if (symbol>=0x20 && symbol<=0x3F) myIndex = symbol-0x20;
				break;
				#endif
				
			#ifdef FONT_UA_MAP
			case fcstEngUaSet: 
				//  - , .  +  (.  .)
				if (symbol=='') myIndex=0x60;
				else if (symbol=='') myIndex=0x61; 
				else if (symbol=='') myIndex=0x62;
				else if (symbol=='') myIndex=0x63;
				else if (symbol=='') myIndex=0x64;
				else if (symbol=='') myIndex=0x65;
				else if (symbol=='') myIndex=0x66;
				else if (symbol=='') myIndex=0x67;
				else if (symbol>=0xC0) myIndex=symbol-0x58;
				else goto engCheck;
				break;
				#endif
				
			#ifdef FONT_RU_MAP
			case fcstEngRusSet: 
				//  - , .  .  (.  .)
				if (symbol>=0xC0) myIndex=symbol-0x60;
				else goto engCheck;
				break;
				#endif
	
			case fcstEngSet:
			default: 


#if defined(FONT_RU_MAP) || defined(FONT_UA_MAP)
	engCheck:
#endif
				//  - ,   (.  .)
				//    ,      
				if (symbol>=0x20 && symbol<=0x7F) myIndex = symbol-0x20;
				break;
				
			#ifdef FONT_ALL_MAP
			case fcstFullMapSet: 
				//  -    256 
				myIndex = symbol;
				break;
				#endif
		} // switch
		// myIndex -     .	
		//        ,   
		
		switch (currentFontSettings.bitDir) {
			#ifdef FONT_LEFT_MSB
			case bdLeftMsb:
				myIndex *= currentFontSettings.symWidth * currentFontSettings.dataSize;
				break;
				#endif
		
			#ifdef FONT_BOTTOM_MSB
				case bdBottomMsb:
				#endif
			#ifdef FONT_TOP_MSB
				case bdTopMsb:
				#endif
			#if defined(FONT_BOTTOM_MSB) || defined(FONT_TOP_MSB)
				myIndex *= currentFontSettings.symHeight * currentFontSettings.dataSize;
				break;
				#endif
				
			#ifdef FONT_BIT_FLOW
				case bdBitFlow:
				myIndex *= ((unsigned int)currentFontSettings.symHeight * currentFontSettings.symWidth + 7) / 8;
				break;
				#endif
	}	// switch

		
		
// 		if (currentFontSettings.bitDir) myIndex *= currentFontSettings.symWidth; //   -   
// 		else myIndex *= currentFontSettings.symHeight; //   -   
// 		myIndex *= currentFontSettings.dataSize; //      
		
		void* fontPtr = currentFontSettings.font+myIndex;
		
		#if defined(FONT_BLOCK_SIZE_4) || defined(FONT_BLOCK_SIZE_3)
			uint32_t buf = 0;
		#elif defined(FONT_BLOCK_SIZE_2)
			unsigned int buf = 0;
		#else 
			unsigned char buf = 0;
		#endif
		
		unsigned char i,j;	 
		
		#ifdef NO_FONT_EFFECTS 
			#ifdef DISPLAY_USE_AREA
				displaySetArea(x, y, x+currentFontSettings.symWidth-1, y+currentFontSettings.symHeight-1);
			#endif
		#endif
		
		switch (currentFontSettings.bitDir)	 {
			#ifdef FONT_LEFT_MSB
			case bdLeftMsb:
				j=0;
				while (j<currentFontSettings.symHeight) {
					//  
					uint8_t fontBytesCounter = currentFontSettings.dataSize;
					buf = 0;
					while (fontBytesCounter) {
						fontBytesCounter--;
						buf = (buf << 8) + pgm_read_byte(fontPtr++);
					}
					i = currentFontSettings.symWidth;
					while (i) {
						i--;
						#ifdef NO_FONT_EFFECTS
							if (buf & 0x01) specialPutPixel(i,j,x,y,drawColor);
							else specialPutPixel(i,j,x,y,bgColor);
						#else // ifdef NO_FONT_EFFECTS
							// not ifdef NO_FONT_EFFECTS
							unsigned int xColor = drawColor;
							if (currentFontSettings.fontColorMap==fcmUseRowsColorMap && fontColorMapArr) xColor = fontColorMapArr[j];
							if (buf & 0x01) specialPutPixel(i,j,x,y,xColor);
							else if (currentFontSettings.textOutMode==tomNormal) specialPutPixel(i,j,x,y,bgColor);
						#endif // ifdef NO_FONT_EFFECTS
						buf /= 2;
					} // while i
					j++;//y++;
				}// while j
				break;
				#endif
			#if defined(FONT_BOTTOM_MSB)
			case bdBottomMsb:
				#endif
			#if defined(FONT_TOP_MSB)
			case bdTopMsb:
				#endif
			#if defined(FONT_TOP_MSB) || defined(FONT_BOTTOM_MSB)
				i = 0;
				while (i<currentFontSettings.symWidth) {
					//  
					uint8_t fontBytesCounter = currentFontSettings.dataSize;
					buf = 0;
					while (fontBytesCounter) {
						fontBytesCounter--;
						buf = (buf << 8) + pgm_read_byte(fontPtr++);
					}
					j = currentFontSettings.symHeight;
					while (j) {
						j--;
						coordType y1;
						//-----    -       ,     
						#if defined(FONT_TOP_MSB) && defined(FONT_BOTTOM_MSB)
						if (currentFontSettings.bitDir==bdBottomMsb)
						#endif
						#if  defined(FONT_BOTTOM_MSB)
							y1 = currentFontSettings.symHeight-1-j;
						#endif
						#if defined(FONT_TOP_MSB) && defined(FONT_BOTTOM_MSB)
						else
						#endif
						#if defined(FONT_TOP_MSB)
							y1 = j;
						#endif
						//-----------
						#ifdef NO_FONT_EFFECTS
							if (buf & 0x01) specialPutPixel(i,y1,x,y,drawColor);
							else specialPutPixel(i,y1,x,y,bgColor);
						#else // ifdef NO_FONT_EFFECTS
							// not ifdef NO_FONT_EFFECTS
							unsigned int xColor = drawColor;
							if (currentFontSettings.fontColorMap==fcmUseRowsColorMap && fontColorMapArr) xColor = fontColorMapArr[y1];
							if ( buf & 0x01 ) specialPutPixel(i,y1,x,y,xColor);
							else if (currentFontSettings.textOutMode==tomNormal) specialPutPixel(i,y1,x,y,bgColor);
						#endif // ifdef NO_FONT_EFFECTS	
						buf /= 2;
					}	// while j
					i++;
				} // while i
				break;
				#endif
			#ifdef FONT_BIT_FLOW
			case bdBitFlow:
				{
					unsigned char bitMask = 0;
					for (j=0; j<currentFontSettings.symHeight; j++) {
						for (i=0; i<currentFontSettings.symWidth; i++) {
							if (!bitMask) {
								buf = pgm_read_byte(fontPtr++);
								bitMask = 0x80;
							}
							#ifdef NO_FONT_EFFECTS
								if (buf & bitMask) specialPutPixel(i,j,x,y,drawColor);
								else specialPutPixel(i,j,x,y,bgColor);
							#else // ifdef NO_FONT_EFFECTS
								// not ifdef NO_FONT_EFFECTS
								unsigned int xColor = drawColor;
								if (currentFontSettings.fontColorMap==fcmUseRowsColorMap && fontColorMapArr) xColor = fontColorMapArr[j];
								if (buf & bitMask) specialPutPixel(i,j,x,y,xColor);
								else if (currentFontSettings.textOutMode==tomNormal) specialPutPixel(i,j,x,y,bgColor);
							#endif // ifdef NO_FONT_EFFECTS
							bitMask >>= 1;
						}	//for i
					}	// for j
				}
				break;	
				#endif
		}// switch
	#ifndef NO_FONT_EFFECTS
		if (currentFontSettings.textIntervalFill && currentFontSettings.textOutMode==tomNormal){
	#else
		if (currentFontSettings.textIntervalFill /*&& currentFontSettings.textOutMode==tomNormal*/){
		#ifdef DISPLAY_USE_AREA
			if (currentFontSettings.textIntervalFill == tifFillLeft) displaySetArea(x-currentFontSettings.symbolsInterval, y, x-1, y+currentFontSettings.symHeight-1);
			else displaySetArea(x+currentFontSettings.symWidth, y, x+currentFontSettings.symWidth+currentFontSettings.symbolsInterval-1, y+currentFontSettings.symHeight-1);
		#endif
	#endif
			i = 0;
			while (i < currentFontSettings.symbolsInterval) {
				for (j = 0; j < currentFontSettings.symHeight; j++) {
					if (currentFontSettings.textIntervalFill == tifFillLeft) specialPutPixel(-1 * ((int)i+1), j, x, y,bgColor );
					else specialPutPixel(currentFontSettings.symWidth + i, j, x, y, bgColor);
				} // for j
				i++;
			} //while (i 
		}  //if (currentFontSettings.textIntervalFill)
	}
	 
	void specialPutString(char isPGM, coordType x, coordType y, const char* strArr, displayFontType fontType)	{
		while (1)	{
			char ch=0;
			if (isPGM) ch = pgm_read_byte(strArr);
			else ch = *strArr;
			if (!ch) return;
			putSymbol(x,y,ch,fontType);
			strArr++;
			int step = currentFontSettings.symWidth + currentFontSettings.symbolsInterval;
			#ifndef NO_FONT_EFFECTS
				if (currentFontSettings.textMirror == tmHMirror) step=-step;
				switch (currentFontSettings.textOrientation) {
					case toNormal:
						x += step;
						break;
					case to90:
						y -= step;
						break;
					case to180:
						x -= step;
						break;
					case to270:
						y += step;
						break;
				} // switch
			#else 
				x += step;
			#endif
		}
	}
	 
	void putString  (coordType x, coordType y, char strArr[], displayFontType fontType){
		specialPutString(0, x, y, strArr, fontType);
	}
	
	void putFString (coordType x, coordType y, const char* fStrArr, displayFontType fontType){
		specialPutString(1, x, y, fStrArr, fontType);		
	}
	
#endif	  //   #ifdef DISPLAY_USE_FONTS

// 	************

unsigned char getBit(unsigned char** ptr, unsigned char init){
	static unsigned char mask = 0;
	static unsigned char data = 0;
	if (init) {
		mask = 0;
		data = 0;
		return 0;
	}
	if (!mask) {
		data = pgm_read_byte(*ptr);
		(*ptr)++;
		mask = 0x80;
	}
	unsigned char result = (data & mask) ? 1 : 0;
	mask >>= 1;
	return result;
}

void putFSprite(coordType x, coordType y, const spriteType* sprite){
	unsigned char* dataPtr = (void*)sprite;
	#ifndef DISPLAY_SHORT_COORDINATES
		dataPtr+= 4 + 1; //     +    
	#else
		dataPtr+= 2 + 1; //     +    
	#endif
	unsigned int customPalette[4];
	customPalette[0] = pgm_read_word(dataPtr);		dataPtr+=2;
	customPalette[1] = pgm_read_word(dataPtr);		dataPtr+=2;
	customPalette[2] = pgm_read_word(dataPtr);		dataPtr+=2;
	customPalette[3] = pgm_read_word(dataPtr);
	putFSpriteCustom(x, y, sprite, customPalette);
}

void putFSpriteCustom(coordType x, coordType y, const spriteType* sprite, unsigned int customPalette[]){
	unsigned char* dataPtr = (void*)sprite;
	#ifndef DISPLAY_SHORT_COORDINATES
		coordType width  = pgm_read_word(dataPtr);		dataPtr+=2;
		coordType height = pgm_read_word(dataPtr);		dataPtr+=2;
	#else
		coordType width  = pgm_read_byte(dataPtr++);		
		coordType height = pgm_read_byte(dataPtr++);		
	#endif
	bitsPerPixelType bpp = pgm_read_byte(dataPtr++);
	dataPtr += 8; //  
	//unsigned int customPalette[4];
	//customPalette[0] = pgm_read_word(dataPtr);		dataPtr+=2;
	//customPalette[1] = pgm_read_word(dataPtr);		dataPtr+=2;
	//customPalette[2] = pgm_read_word(dataPtr);		dataPtr+=2;
	//customPalette[3] = pgm_read_word(dataPtr);		dataPtr+=2;
	//      

	coordType iw = 0, ih = 0;
	#ifdef DISPLAY_USE_AREA
		displaySetArea(x, y, x+width-1, y+height-1);
	#endif
	getBit(0,1);//init bitflow
	while ( 1 ) {
		//    -  
		if (getBit(&dataPtr,0)) {
			//   
			unsigned char sz = 0, i;
			for (i=0; i<6; i++) sz = (sz << 1) | getBit(&dataPtr,0);
			sz += 2;
			// sz -   
			unsigned char iColor;
			iColor = getBit(&dataPtr,0);
			if (bpp == bpp2bits) iColor = (iColor << 1) | getBit(&dataPtr,0);
			// iColor -   
			unsigned int c = customPalette[iColor]; // 
			while(sz) {
				sz--;
				#ifdef DISPLAY_USE_AREA
					displaySendColor(c);
				#else
					displayPutPixel(x+iw, y+ih, c);
				#endif
				iw++;
				if (iw == width) {
					iw = 0;
					ih++;
					if (ih==height) return;
				}
			}// while sz
		} else {
			//     
			unsigned char sz = 0, i;
			for (i=0; i<4; i++) sz = (sz<<1) | getBit(&dataPtr,0);
			sz += 1;
			// sz -    
			while(sz) {
				sz--;
				unsigned char iColor;
				iColor = getBit(&dataPtr,0);
				if (bpp == bpp2bits) iColor = (iColor << 1) | getBit(&dataPtr,0);
				// iColor -   
				#ifdef DISPLAY_USE_AREA
					displaySendColor(customPalette[iColor]);
				#else
					displayPutPixel(x+iw, y+ih, customPalette[iColor]);
				#endif
				iw++;
				if (iw == width) {
					iw = 0;
					ih++;
					if (ih==height) return;
				}
			}// while sz
		}
	}  // while (1);
	
	//#undef GETBIT
}

/*****************************************************************************
	 RLE:          
	       .
	        .
	          .
	       ,        .
	 :
	      -    
	1 - , 0 - 
	 -  .
	      6 
	  ,      2  , 6  (0..63)   
	   2  65 
	    1  2     (     ).
	,   8    9    .
	
	  -     4 
	  ,         1 , 4  (0..15) 
	    1  16 .
	        .        (     ).
	,    6  21       7  37   .

*****************************************************************************/


