/*
 * i2c_mstr.h
 *
 * Created: 15.03.2019 15:31:16
 *  Author: GoldenAndy
 */ 


/*
 *	   i2c   
 */

/************************************
 *  [write] & [read] -   -   
 *
 *	  
 *	----------------------------------
 *	
 *	  + [read]
 *	    + ACK
 *	    + ACK
 *	    + ACK
 *	..
 *	    + NACK
 *	
 * 
 *	   
 *	--------------------------------------
 *	
 *	 [write]
 *	  ,    (  EEPROM,    ..)
 *	 ,
 *	  [read],
 *	    + ACK
 *	    + ACK
 *	    + ACK
 *	..
 *	    + NACK
 *	C
 *
 *	 
 *	------------------------
 *	
 *	  [write]
 *	 
 *	 
 *	 
 *	..
 *	
 *
 ************************************/



#ifndef I2C_MSTR_H_
#define I2C_MSTR_H_

	#include <avr/io.h>
	#include <stdint.h>
	
	typedef enum {
		iicError	=	0x00,
		iicOk		=	0x01,
		iicOkACK	=	0x03,
		iicOkNACK	=	0x05
		} iicResultType;
	
	typedef enum {
		iicWriteData,
		iicReadData 
		} iicDataDirectionType;
		
	typedef enum {
		iicNACKflag,
		iicACKflag
		} iicAckFlagType;
	
	//  ,   ,      STOP,
	//     - iicError     iicOk, iicOkACK  iicOkNACK

	//	   i2c.
	//	   i2c-  !
	void iicInit(uint8_t TWBR_val, uint8_t TWSR_val);				
	
	//	    																
	iicResultType iicSetStart(void);								 
	
	//	    i2c
	//   -  8 ,  
	//	    dataDirection
	iicResultType iicSendAddr(uint8_t iicAddr_8bit, iicDataDirectionType dataDirection);	
																	
	//	    i2c																
	iicResultType iicSendData(uint8_t data);
	
	//	    i2c						
	iicResultType iicReceiveData(iicAckFlagType ackFlag);			
	
	//	   i2c 
	uint8_t iisGetReceivedData(void);	
	
	//	    							
	void iicSetStop(void);											

	//     (1 ),  .  1   , 0  
	uint8_t iic_writeAddrWriteData(uint8_t i2cAddr8bit, uint8_t regAddr, uint8_t* buffer, uint8_t dataCount);
	
	//     (1 ),  .  1   , 0  
	uint8_t iic_writeAddrReadData(uint8_t i2cAddr8bit, uint8_t Reg, uint8_t* buffer, uint8_t dataCount);
	
#endif /* I2C_MSTR_H_ */


