#include "keyAntiBounce.h"

#ifdef _KEY_BOUNCE_8BIT_
	#define  MID_VAL		0x3F
	#define  FLAG_MASK		0x80
	#define  COUNTER_MASK	0x7F
#endif

#ifdef _KEY_BOUNCE_16BIT_
	#define	 MID_VAL		0x3FFF
	#define  FLAG_MASK		0x8000
	#define  COUNTER_MASK	0x7FFF
#endif
  
	/********************
	 *	name		:	keyAntiBounce
	 *	description	:	  
	 *	arguments	:	keyCounter			->     
	 *					isPhysicallyPressed ->   
	 *					keyBounceValue		->       
	 *	result		:	1 -  , 0 -  
	 *	
	 *	   ,    -     1,   - .
	 *	      -       
	 *	 1 -   .
	 *	    ,         . 
	 *	      0 -   .
	 *	,        (     
	 *	   )
	 *     ,     (_KEY_BOUNCE_8BIT_  	_KEY_BOUNCE_16BIT_), 
	 *	  keyCounter     keyBounceValue  8-  16- .
	 *	          2..126  8- ,  2..32766 -  16-.
	 *	(        (8  16   .    
	 *	        )
	 *	  -    (true - )
	 *	
	 *	 :   -   ,   -  .
	 *	  . 
	 *	         (MID_VAL+keyBounceValue/2) -  ,
	 *		 -    .
	 *	         (MID_VAL-keyBounceValue/2) -  , 
	 *		 -    .
	 *	 MID_VAL -   ,       .
	 *	 8-   0x3F,  16- - 0x3FFF
	 *	       ,   .
	 ********************/

uint8_t keyAntiBounce (bounceCounterType * keyCounter, uint8_t isPhysicallyPressed, bounceCounterType keyBounceValue) {
	keyBounceValue /= 2;
	keyBounceValue &= MID_VAL;
		
	bounceCounterType flagPress = (*keyCounter) & FLAG_MASK;	//    -   
	bounceCounterType kCtr = *keyCounter & COUNTER_MASK;		//     

	if (! isPhysicallyPressed) {
		//   
		if (kCtr > MID_VAL-keyBounceValue) kCtr--;
		else {
			kCtr = MID_VAL-keyBounceValue;
			flagPress = 0;
		}
	} else {
		//   
		if (kCtr < MID_VAL+keyBounceValue) (kCtr)++;
		else {
			kCtr = MID_VAL+keyBounceValue;
			flagPress = FLAG_MASK;
		}		
	}
	*keyCounter = kCtr | flagPress;
	return flagPress != 0;		
}


#ifdef	_KEY_BOUNCE_INDIVIDUAL_TIMINGS_
void keyProcessingInit(keyProcessingStruct * keyStruct, uint8_t keyMode, uint16_t longTime, uint16_t repeatTime, bounceCounterType bounceTime) {
	keyStruct->bounceTime = bounceTime;
	keyStruct->longPressTime = longTime;
	keyStruct->repeatPressTime = repeatTime<longTime ? repeatTime : longTime;
#else	//_KEY_BOUNCE_INDIVIDUAL_TIMINGS_ is not defined
void keyProcessingInit(keyProcessingStruct * keyStruct, uint8_t keyMode) {
#endif //ifdef _KEY_BOUNCE_INDIVIDUAL_TIMINGS_
	keyStruct->bounceCounter = MID_VAL;
	keyStruct->pressCounter = 0;
	keyStruct->previousState = 0;
	if (keyMode & _KEY_FLAG_REPEAT_) keyMode |= _KEY_FLAG_SHORTPRESS_;
	if ((keyMode & (_KEY_FLAG_LONGPRESS_ | _KEY_FLAG_REPEAT_)) == (_KEY_FLAG_LONGPRESS_ | _KEY_FLAG_REPEAT_)) keyMode &= ~(_KEY_FLAG_LONGPRESS_ | _KEY_FLAG_REPEAT_);
	keyStruct->processingFlags = keyMode;
}


uint8_t keyProcessing(keyProcessingStruct * keyStruct, uint8_t isPhysicallyPressed, uint8_t forceReset){
	uint8_t result = 0;
	if (forceReset){
		keyStruct->bounceCounter = MID_VAL;
		keyStruct->pressCounter = 0;
		keyStruct->previousState = 0;
		return result;
	}
	//  
	#ifdef	_KEY_BOUNCE_INDIVIDUAL_TIMINGS_
		uint8_t onPressed = keyAntiBounce( &keyStruct->bounceCounter, isPhysicallyPressed, keyStruct->bounceTime );
	#else
		uint8_t onPressed = keyAntiBounce( &keyStruct->bounceCounter, isPhysicallyPressed, _KEY_BOUNCE_TIME_ );
	#endif
	if (keyStruct->processingFlags == 0) return 0;
	//    
	if (!(keyStruct->previousState) && onPressed) {
		if (keyStruct->processingFlags & _KEY_FLAG_EVENTS_) result |= _KEY_EVENT_ON_PRESS_;
		keyStruct->previousState = 1;
		keyStruct->pressCounter = 0;
		if ((keyStruct->processingFlags & (_KEY_FLAG_SHORTPRESS_ | _KEY_FLAG_LONGPRESS_)) == _KEY_FLAG_SHORTPRESS_){
			//         -      
			return result | _KEY_EVENT_SHORTPRESS_;
		} else return result;
	}	//    
	
	//    
	if (keyStruct->previousState && !onPressed) {
		if (keyStruct->processingFlags & _KEY_FLAG_EVENTS_) result |= _KEY_EVENT_ON_RELEASE_;
		keyStruct->previousState = 0;
		if (keyStruct->processingFlags & _KEY_FLAG_LONGPRESS_)	 {
			#ifdef	_KEY_BOUNCE_INDIVIDUAL_TIMINGS_
				uint16_t longTime = keyStruct->longPressTime;
			#else
				uint16_t longTime = (_KEY_LONG_PRESS_TIME_);
			#endif
			if (keyStruct->pressCounter < longTime) {	 //     
				if (keyStruct->processingFlags & _KEY_FLAG_SHORTPRESS_) return result | _KEY_EVENT_SHORTPRESS_;	//     
				else return result;
			} else return result; //      .
		//  
		} else return 0;  //   ,  ,     
	}	//    
	
	//    
	if (keyStruct->previousState && onPressed) {
		if ( !(keyStruct->processingFlags & (_KEY_FLAG_LONGPRESS_ | _KEY_FLAG_REPEAT_) )) return result; //       -    
		#ifdef	_KEY_BOUNCE_INDIVIDUAL_TIMINGS_
			uint16_t longTime = keyStruct->longPressTime,
					repeatTime = keyStruct->repeatPressTime;
		#else
			uint16_t longTime = (_KEY_LONG_PRESS_TIME_),
					repeatTime = (_KEY_REPEAT_TIME_) < (_KEY_LONG_PRESS_TIME_) ? (_KEY_REPEAT_TIME_) : (_KEY_LONG_PRESS_TIME_);
		#endif
		if ( keyStruct->pressCounter < 0xFFFF ) keyStruct->pressCounter++;
		if ( keyStruct->pressCounter == longTime ) {
			if ( keyStruct->processingFlags & _KEY_FLAG_LONGPRESS_ )  return result | _KEY_EVENT_LONGPRESS_;
			if ( keyStruct->processingFlags & _KEY_FLAG_REPEAT_ ) {
				keyStruct->pressCounter -= repeatTime;
				return result | _KEY_EVENT_REPEAT_ | _KEY_EVENT_SHORTPRESS_;
			}
			return result;
		} else return result;
	}
	//  .
	return result;
}


#ifdef	_KEY_BOUNCE_INDIVIDUAL_TIMINGS_
void encoderProcessingInit(encoderProcessingStruct * encStruct,
	bounceCounterType rotaryBounceTime,
	uint8_t keyMode, uint16_t longTime, uint16_t repeatTime, bounceCounterType keyBounceTime){
	encStruct->rotaryBounce = rotaryBounceTime;
	encStruct->btn.bounceTime = keyBounceTime;
	encStruct->btn.longPressTime = longTime;
	encStruct->btn.repeatPressTime = repeatTime<longTime ? repeatTime : longTime;
#else	//_KEY_BOUNCE_INDIVIDUAL_TIMINGS_ is not defined
void encoderProcessingInit(encoderProcessingStruct * encStruct, uint8_t keyMode){
	encStruct->rotaryBounce = _KEY_BOUNCE_TIME_;
	#endif
	encStruct->chA = MID_VAL;
	encStruct->chB = MID_VAL;
	encStruct->prevEncValue = 0;
	encStruct->rotates = 0;

	encStruct->btn.bounceCounter = MID_VAL;
	encStruct->btn.pressCounter = 0;
	encStruct->btn.previousState = 0;
	if (keyMode & _KEY_FLAG_REPEAT_) keyMode |= _KEY_FLAG_SHORTPRESS_;
	if ((keyMode & (_KEY_FLAG_LONGPRESS_ | _KEY_FLAG_REPEAT_)) == (_KEY_FLAG_LONGPRESS_ | _KEY_FLAG_REPEAT_)) keyMode &= ~(_KEY_FLAG_LONGPRESS_ | _KEY_FLAG_REPEAT_);
	encStruct->btn.processingFlags = keyMode;
}
	

uint8_t encoderProcessing(encoderProcessingStruct * encStruct, uint8_t isPhysicallyPressed, uint8_t isPhysicallyChA, uint8_t isPhysicallyChB, uint8_t forceReset){
	uint8_t result = 0;
	if (forceReset){
		encStruct->btn.bounceCounter = MID_VAL;
		encStruct->btn.pressCounter = 0;
		encStruct->btn.previousState = 0;
		encStruct->chA = MID_VAL;
		encStruct->chB = MID_VAL;
		encStruct->prevEncValue = 0;
		encStruct->rotates = 0;
		return result;
	}
	#define CH_A	0x01
	#define CH_B	0x02
	
	uint8_t encChannels = 0, rFlag = 0;
	if (keyAntiBounce(&encStruct->chA, isPhysicallyChA, encStruct->rotaryBounce)) encChannels |= CH_A; 
	if (keyAntiBounce(&encStruct->chB, isPhysicallyChB, encStruct->rotaryBounce)) encChannels |= CH_B; 
	
	if (encStruct->prevEncValue != encChannels) {
		if (encStruct->prevEncValue == (CH_A | CH_B)) {
			if (encChannels == CH_A && encStruct->rotates < 127) { encStruct->rotates += 1; rFlag = 1; }
			if (encChannels == CH_B && encStruct->rotates >-127) { encStruct->rotates -= 1; rFlag = 1; }
		}
		encStruct->prevEncValue = encChannels;
	}
	if (encStruct->btn.processingFlags & _KEY_ENCODER_NO_PRESS_ON_ROTATE_ && rFlag){
		result = keyProcessing(&encStruct->btn, 0, 1);
	} else result = keyProcessing(&encStruct->btn, isPhysicallyPressed,0);
	if (rFlag) result |= _KEY_EVENT_ROTATE_;
	return result;
	
	#undef CH_A
	#undef CH_B
}

int8_t encoderGetRotates(encoderProcessingStruct * encStruct){
	int8_t result = encStruct->rotates;
	encStruct->rotates = 0;
	return result;
}
	  

#undef	MID_VAL
#undef  FLAG_MASK
#undef  COUNTER_MASK