/*
 * systemClock.c
 *
 * Created: 28.08.2018 15:05:00
 *  Author: GoldenAndy
 */ 

#include "systemClock.h"

#if !defined(SYSTEMCLOCK_INTERRUPT_VECTOR) && !defined(SYSTEMCLOCK_USE_EXTERNAL_CALL)
#error # define SYSTEMCLOCK_INTERRUPT_VECTOR or SYSTMCLOCK_USE_EXTERNAL_CALL not found.
#endif

#define rf_NoReset			0
#define rf_ResetCounter		1
#define rf_ResetPrevious	2

volatile static uint8_t resetFlag = 0;
volatile static systemClockCounterType sysCounter = 0;

#ifdef SYSTEMCLOCK_INTERRUPT_VECTOR
ISR(SYSTEMCLOCK_INTERRUPT_VECTOR){
#endif
#ifdef SYSTEMCLOCK_USE_EXTERNAL_CALL
void systemClockTick(void){
#endif
	sysCounter++;
	if (resetFlag == rf_ResetCounter) {
		sysCounter = 0;
		resetFlag = rf_ResetPrevious;
	}
}

systemClockCounterType getSysClock(void){
	systemClockCounterType c1,c2;
	do {
		c1 = sysCounter;
		c2 = sysCounter;
	} while (c1 != c2);
	return c1;
}

systemClockCounterType getSysClockInterval(void){
	static systemClockCounterType previousCounter = 0;
	if (resetFlag == rf_ResetPrevious) {
		previousCounter = 0;
		resetFlag = rf_NoReset;
	}
	systemClockCounterType delta = getSysClock() - previousCounter;
	previousCounter += delta;
	return delta;
}

void resetSysClock(void){
	resetFlag = rf_ResetCounter;
}

#undef rf_NoReset
#undef rf_ResetCounter
#undef rf_ResetPrevious
